"""Core module control interfaces."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, Generic, List, Optional, Sequence, TypeVar

from .labware import AbstractLabware, LabwareCoreType
from .tasks import AbstractTaskCore
from opentrons.drivers.types import (
    HeaterShakerLabwareLatchStatus,
    ThermocyclerLidStatus,
)
from opentrons.hardware_control.modules.types import (
    MagneticStatus,
    ModuleModel,
    ModuleType,
    SpeedStatus,
    TemperatureStatus,
    ThermocyclerStep,
)
from opentrons.protocol_engine.types import ABSMeasureMode
from opentrons.types import DeckSlotName


class AbstractModuleCore(ABC, Generic[LabwareCoreType]):
    """Abstract core module control interface."""

    MODULE_TYPE: ClassVar[ModuleType]

    @abstractmethod
    def get_model(self) -> ModuleModel:
        """Get the module's model identifier."""

    @abstractmethod
    def get_deck_slot(self) -> DeckSlotName:
        """Get the module's deck slot."""

    @abstractmethod
    def get_deck_slot_id(self) -> str:
        """Get the module's deck slot in a robot accurate format."""

    @abstractmethod
    def get_display_name(self) -> str:
        """Get the module's display name."""


ModuleCoreType = TypeVar("ModuleCoreType", bound=AbstractModuleCore[Any])


class AbstractTemperatureModuleCore(
    AbstractModuleCore[LabwareCoreType], Generic[LabwareCoreType]
):
    """Core control interface for an attached Temperature Module."""

    MODULE_TYPE: ClassVar = ModuleType.TEMPERATURE

    @abstractmethod
    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""

    @abstractmethod
    def set_target_temperature(self, celsius: float) -> AbstractTaskCore:
        """Set the Temperature Module's target temperature in °C."""

    @abstractmethod
    def wait_for_target_temperature(self, celsius: Optional[float] = None) -> None:
        """Wait until the module's target temperature is reached.

        Specifying a value for ``celsius`` that is different than
        the module's current target temperature may behave unpredictably.
        """

    @abstractmethod
    def deactivate(self) -> None:
        """Deactivate the Temperature Module."""

    @abstractmethod
    def get_current_temperature(self) -> float:
        """Get the module's current temperature in °C."""

    @abstractmethod
    def get_target_temperature(self) -> Optional[float]:
        """Get the module's target temperature in °C, if set."""

    @abstractmethod
    def get_status(self) -> TemperatureStatus:
        """Get the module's current temperature status."""


class AbstractMagneticModuleCore(
    AbstractModuleCore[LabwareCoreType], Generic[LabwareCoreType]
):
    """Core control interface for an attached Magnetic Module."""

    MODULE_TYPE: ClassVar = ModuleType.MAGNETIC

    @abstractmethod
    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""

    @abstractmethod
    def engage(
        self,
        height_from_base: Optional[float] = None,
        height_from_home: Optional[float] = None,
    ) -> None:
        """Raise the module's magnets.

        Only one of `height_from_base` or `height_from_home` may be specified.

        Args:
            height_from_base: Distance from labware base to raise the magnets.
            height_from_home: Distance from motor home position to raise the magnets.
        """

    @abstractmethod
    def engage_to_labware(
        self, offset: float = 0, preserve_half_mm: bool = False
    ) -> None:
        """Raise the module's magnets up to its loaded labware.

        Args:
            offset: Offset from the labware's default engage height.
            preserve_half_mm: For labware whose definitions
                erroneously use half-mm for their defined default engage height,
                use the value directly instead of converting it to real millimeters.

        Raises:
            Exception: Labware is not loaded or has no default engage height.
        """

    @abstractmethod
    def disengage(self) -> None:
        """Lower the magnets back into the module."""

    @abstractmethod
    def get_status(self) -> MagneticStatus:
        """Get the module's current magnet status."""


class AbstractThermocyclerCore(
    AbstractModuleCore[LabwareCoreType], Generic[LabwareCoreType]
):
    """Core control interface for an attached Thermocycler Module."""

    MODULE_TYPE: ClassVar = ModuleType.THERMOCYCLER

    @abstractmethod
    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""

    @abstractmethod
    def open_lid(self) -> ThermocyclerLidStatus:
        """Open the Thermocycler's lid."""

    @abstractmethod
    def close_lid(self) -> ThermocyclerLidStatus:
        """Close the Thermocycler's lid."""

    @abstractmethod
    def set_target_block_temperature(
        self,
        celsius: float,
        ramp_rate: Optional[float],
        hold_time_seconds: Optional[float] = None,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """Set the target temperature for the well block, in °C.

        Note:
            If ``hold_time_seconds`` is not specified, the Thermocycler
            will proceed to the next command after ``temperature`` is reached.
        Args:
            celsius: The target temperature, in °C.
            hold_time_seconds: The number of seconds to hold, after reaching
                ``temperature``, before proceeding to the next command.
            block_max_volume: The maximum volume of any individual well
                of the loaded labware. If not supplied, the thermocycler
                will default to 25µL/well.
        """

    @abstractmethod
    def start_set_target_block_temperature(
        self,
        celsius: float,
        ramp_rate: Optional[float],
        block_max_volume: Optional[float] = None,
    ) -> AbstractTaskCore:
        """Start setting the target temperature for the well block, in °C.

        Note:
            If ``hold_time_seconds`` is not specified, the Thermocycler
            will proceed to the next command after ``temperature`` is reached.
        Args:
            celsius: The target temperature, in °C.
            block_max_volume: The maximum volume of any individual well
                of the loaded labware. If not supplied, the thermocycler
                will default to 25µL/well.
        """

    @abstractmethod
    def wait_for_block_temperature(self) -> None:
        """Wait for target block temperature to be reached."""

    @abstractmethod
    def set_target_lid_temperature(self, celsius: float) -> None:
        """Set the target temperature for the heated lid, in °C."""

    @abstractmethod
    def start_set_target_lid_temperature(self, celsius: float) -> AbstractTaskCore:
        """Start setting the target temperature for the heated lid, in °C."""

    @abstractmethod
    def wait_for_lid_temperature(self) -> None:
        """Wait for target lid temperature to be reached."""

    @abstractmethod
    def execute_profile(
        self,
        steps: List[ThermocyclerStep],
        repetitions: int,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """Execute a Thermocycler Profile.

        Profile defined as a cycle of ``steps`` to repeat for a given number of ``repetitions``

        Note:
            Unlike the :py:meth:`set_block_temperature`, either or both of
            'hold_time_minutes' and 'hold_time_seconds' must be defined
            and finite for each step.
        Args:
            steps: List of unique steps that make up a single cycle.
                Each list item should be a dictionary that maps to
                the parameters of the :py:meth:`set_block_temperature`
                method with keys 'temperature', 'hold_time_seconds',
                and 'hold_time_minutes'.
            repetitions: The number of times to repeat the cycled steps.
            block_max_volume: The maximum volume of any individual well
                of the loaded labware. If not supplied, the thermocycler
                will default to 25µL/well.
        """

    @abstractmethod
    def start_execute_profile(
        self,
        steps: List[ThermocyclerStep],
        repetitions: int,
        block_max_volume: Optional[float] = None,
    ) -> AbstractTaskCore:
        """Start a Thermocycler Profile.

        Profile defined as a cycle of ``steps`` to repeat for a given number of ``repetitions``

        Note:
            Unlike the :py:meth:`execute_profile`, once the profile has started
            the protocol will immediately move on to the next command, rather than waiting
            for it to finish.
        Args:
            steps: List of unique steps that make up a single cycle.
                Each list item should be a dictionary that maps to
                the parameters of the :py:meth:`set_block_temperature`
                method with keys 'temperature', 'hold_time_seconds',
                and 'hold_time_minutes'.
            repetitions: The number of times to repeat the cycled steps.
            block_max_volume: The maximum volume of any individual well
                of the loaded labware. If not supplied, the thermocycler
                will default to 25µL/well.
        """

    @abstractmethod
    def deactivate_lid(self) -> None:
        """Turn off the heated lid."""

    @abstractmethod
    def deactivate_block(self) -> None:
        """Turn off the well block temperature controller"""

    @abstractmethod
    def deactivate(self) -> None:
        """Turn off the well block temperature controller, and heated lid"""

    @abstractmethod
    def get_lid_position(self) -> Optional[ThermocyclerLidStatus]:
        """Get the Thermocycler's lid position."""

    @abstractmethod
    def get_block_temperature_status(self) -> TemperatureStatus:
        """Get the Thermocycler's block temperature status."""

    @abstractmethod
    def get_lid_temperature_status(self) -> Optional[TemperatureStatus]:
        """Get the Thermocycler's lid temperature status."""

    @abstractmethod
    def get_block_temperature(self) -> Optional[float]:
        """Get the Thermocycler's current block temperature in °C."""

    @abstractmethod
    def get_block_target_temperature(self) -> Optional[float]:
        """Get the Thermocycler's target block temperature in °C."""

    @abstractmethod
    def get_lid_temperature(self) -> Optional[float]:
        """Get the Thermocycler's current lid temperature in °C."""

    @abstractmethod
    def get_lid_target_temperature(self) -> Optional[float]:
        """Get the Thermocycler's target lid temperature in °C."""

    @abstractmethod
    def get_ramp_rate(self) -> Optional[float]:
        """Get the Thermocycler's current ramp rate in °C/sec."""

    @abstractmethod
    def get_hold_time(self) -> Optional[float]:
        """Get the remaining hold time in seconds."""

    @abstractmethod
    def get_total_cycle_count(self) -> Optional[int]:
        """Get number of repetitions for current set cycle."""

    @abstractmethod
    def get_current_cycle_index(self) -> Optional[int]:
        """Get index of the current set cycle repetition."""

    @abstractmethod
    def get_total_step_count(self) -> Optional[int]:
        """Get number of steps within the current cycle."""

    @abstractmethod
    def get_current_step_index(self) -> Optional[int]:
        """Get the index of the current step within the current cycle."""


class AbstractHeaterShakerCore(
    AbstractModuleCore[LabwareCoreType], Generic[LabwareCoreType]
):
    """Core control interface for an attached Heater-Shaker Module."""

    MODULE_TYPE: ClassVar = ModuleType.HEATER_SHAKER

    @abstractmethod
    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""

    @abstractmethod
    def set_target_temperature(self, celsius: float) -> AbstractTaskCore:
        """Set the labware plate's target temperature in °C."""

    @abstractmethod
    def wait_for_target_temperature(self) -> None:
        """Wait for the labware plate's target temperature to be reached."""

    @abstractmethod
    def set_and_wait_for_shake_speed(self, rpm: int) -> None:
        """Set the shaker's target shake speed and wait for it to spin up."""

    @abstractmethod
    def set_shake_speed(self, rpm: int) -> AbstractTaskCore:
        """Set the shaker's target shake speed."""

    @abstractmethod
    def open_labware_latch(self) -> None:
        """Open the labware latch."""

    @abstractmethod
    def close_labware_latch(self) -> None:
        """Close the labware latch."""

    @abstractmethod
    def deactivate_shaker(self) -> None:
        """Stop shaking."""

    @abstractmethod
    def deactivate_heater(self) -> None:
        """Stop heating."""

    @abstractmethod
    def get_current_temperature(self) -> float:
        """Get the labware plate's current temperature in °C."""

    @abstractmethod
    def get_target_temperature(self) -> Optional[float]:
        """Get the labware plate's target temperature in °C, if set."""

    @abstractmethod
    def get_current_speed(self) -> int:
        """Get the shaker's current speed in RPM."""

    @abstractmethod
    def get_target_speed(self) -> Optional[int]:
        """Get the shaker's target speed in RPM, if set."""

    @abstractmethod
    def get_temperature_status(self) -> TemperatureStatus:
        """Get the module's heater status."""

    @abstractmethod
    def get_speed_status(self) -> SpeedStatus:
        """Get the module's heater status."""

    @abstractmethod
    def get_labware_latch_status(self) -> HeaterShakerLabwareLatchStatus:
        """Get the module's labware latch status."""


class AbstractMagneticBlockCore(
    AbstractModuleCore[LabwareCoreType], Generic[LabwareCoreType]
):
    """Core control interface for an attached Magnetic Block."""

    MODULE_TYPE: ClassVar = ModuleType.MAGNETIC_BLOCK


class AbstractAbsorbanceReaderCore(
    AbstractModuleCore[LabwareCoreType], Generic[LabwareCoreType]
):
    """Core control interface for an attached Absorbance Reader Module."""

    MODULE_TYPE: ClassVar = ModuleType.ABSORBANCE_READER

    @abstractmethod
    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""

    @abstractmethod
    def initialize(
        self,
        mode: ABSMeasureMode,
        wavelengths: List[int],
        reference_wavelength: Optional[int] = None,
    ) -> None:
        """Initialize the Absorbance Reader by taking zero reading."""

    @abstractmethod
    def read(self, filename: Optional[str] = None) -> Dict[int, Dict[str, float]]:
        """Get an absorbance reading from the Absorbance Reader."""

    @abstractmethod
    def close_lid(self) -> None:
        """Close the Absorbance Reader's lid."""

    @abstractmethod
    def open_lid(self) -> None:
        """Open the Absorbance Reader's lid."""

    @abstractmethod
    def is_lid_on(self) -> bool:
        """Return True if the Absorbance Reader's lid is currently closed."""


class AbstractFlexStackerCore(
    AbstractModuleCore[LabwareCoreType], Generic[LabwareCoreType]
):
    """Core control interface for an attached Flex Stacker."""

    MODULE_TYPE: ClassVar = ModuleType.FLEX_STACKER

    @abstractmethod
    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""

    @abstractmethod
    def retrieve(self) -> AbstractLabware[Any]:
        """Release a labware from the hopper to the staging slot.

        Returns the retreived primary labware.
        """

    @abstractmethod
    def store(self) -> None:
        """Store a labware in the stacker hopper."""

    @abstractmethod
    def fill(self, count: int | None, message: str | None) -> None:
        """Pause the protocol to allow for filling the stacker."""

    @abstractmethod
    def fill_items(
        self, labware: Sequence[LabwareCoreType], message: str | None
    ) -> None:
        """Pause the protocol to fill with a specific set of labware."""

    @abstractmethod
    def empty(self, message: str | None) -> None:
        """Pause the protocol to allow for emptying the stacker."""

    @abstractmethod
    def set_stored_labware_items(
        self,
        labware: Sequence[LabwareCoreType],
        stacking_offset_z: float | None,
    ) -> None:
        """Configure the stacker to contain a set of labware."""

    @abstractmethod
    def get_max_storable_labware(self) -> int:
        """Get the total number of configured labware the stacker can store."""

    @abstractmethod
    def get_current_storable_labware(self) -> int:
        """Get the amount of space currently available for labware."""

    @abstractmethod
    def get_max_storable_labware_from_list(
        self,
        labware: Sequence[LabwareCoreType],
        overlap_offset: float | None = None,
    ) -> Sequence[LabwareCoreType]:
        """Limit the passed list to how many labware can fit in a stacker."""

    @abstractmethod
    def get_current_storable_labware_from_list(
        self,
        labware: Sequence[LabwareCoreType],
    ) -> Sequence[LabwareCoreType]:
        """Limit the passed list to how many labware can fit in the stacker right now."""

    @abstractmethod
    def get_stored_labware(self) -> Sequence[LabwareCoreType]:
        """Get the currently-stored labware from the stacker."""

    @abstractmethod
    def set_stored_labware(
        self,
        main_load_name: str,
        main_namespace: str | None,
        main_version: int | None,
        lid_load_name: str | None,
        lid_namespace: str | None,
        lid_version: int | None,
        adapter_load_name: str | None,
        adapter_namespace: str | None,
        adapter_version: int | None,
        count: int | None,
        stacking_offset_z: float | None = None,
    ) -> None:
        """Configure the kind of labware that the stacker stores."""


class AbstractVacuumModuleCore(
    AbstractModuleCore[LabwareCoreType], Generic[LabwareCoreType]
):
    """Core control interface for an attached Vacuum Module."""

    MODULE_TYPE: ClassVar = ModuleType.VACUUM_MODULE

    @abstractmethod
    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""



"""Tests for the InstrumentContext public interface."""

import pytest
from decoy import Decoy

from opentrons.protocol_api import MAX_SUPPORTED_VERSION, Labware, Well
from opentrons.protocol_api._liquid import Liquid
from opentrons.protocol_api.core.common import WellCore
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import Location, Point


@pytest.fixture
def mock_well_core(decoy: Decoy) -> WellCore:
    """Get a mock labware implementation core."""
    core = decoy.mock(cls=WellCore)
    decoy.when(core.get_display_name()).then_return("A1 of Cool Labware")
    return core


@pytest.fixture
def api_version() -> APIVersion:
    """Get the API version to test at."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def mock_parent(decoy: Decoy) -> Labware:
    """Get the well's parent."""
    return decoy.mock(cls=Labware)


@pytest.fixture
def subject(
    mock_parent: Labware, mock_well_core: WellCore, api_version: APIVersion
) -> Well:
    """Get a Well test subject with its dependencies mocked out."""
    return Well(parent=mock_parent, core=mock_well_core, api_version=api_version)


@pytest.mark.parametrize("api_version", [APIVersion(2, 13)])
def test_api_version(api_version: APIVersion, subject: Well) -> None:
    """It should have an api_version property."""
    assert subject.api_version == api_version


@pytest.mark.parametrize("api_version", [APIVersion(2, 12)])
def test_api_version_clamped(subject: Well) -> None:
    """It should not allow the API version to go any lower than 2.13."""
    assert subject.api_version == APIVersion(2, 13)


def test_parent(mock_parent: Labware, subject: Well) -> None:
    """It should have a reference to its parent."""
    assert subject.parent is mock_parent


def test_repr(
    decoy: Decoy, mock_well_core: WellCore, mock_parent: Labware, subject: Well
) -> None:
    """It should have a string representation."""
    assert subject.display_name == "A1 of Cool Labware"
    assert repr(subject) == "A1 of Cool Labware"


def test_well_max_volume(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get the well's max volume from the core."""
    decoy.when(mock_well_core.get_max_volume()).then_return(101)
    assert subject.max_volume == 101


def test_well_top(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get a Location representing the top of the well."""
    decoy.when(mock_well_core.get_top(z_offset=4.2)).then_return(Point(1, 2, 3))

    result = subject.top(4.2)

    assert isinstance(result, Location)
    assert result.point == Point(1, 2, 3)
    assert result.labware.as_well() is subject


def test_well_bottom(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get a Location representing the bottom of the well."""
    decoy.when(mock_well_core.get_bottom(z_offset=4.2)).then_return(Point(1, 2, 3))

    result = subject.bottom(4.2)

    assert isinstance(result, Location)
    assert result.point == Point(1, 2, 3)
    assert result.labware.as_well() is subject


def test_well_center(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get a Location representing the center of the well."""
    decoy.when(mock_well_core.get_center()).then_return(Point(1, 2, 3))

    result = subject.center()

    assert isinstance(result, Location)
    assert result.point == Point(1, 2, 3)
    assert result.labware.as_well() is subject


def test_well_meniscus(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get a Location representing the meniscus of the well."""
    result = subject.meniscus(z=4.2, target="end")

    assert isinstance(result, Location)
    assert result.point == Point(0, 0, 4.2)
    assert result.labware.as_well() is subject


def test_has_tip(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get tip state from the core."""
    decoy.when(mock_well_core.has_tip()).then_return(True)
    assert subject.has_tip is True

    decoy.when(mock_well_core.has_tip()).then_return(False)
    assert subject.has_tip is False


def test_load_liquid(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should load a liquid to a location."""
    mocked_liquid = Liquid(
        _id="liquid-id", name="water", description=None, display_color=None
    )

    subject.load_liquid(
        liquid=mocked_liquid,
        volume=20,
    )

    decoy.verify(
        mock_well_core.load_liquid(
            liquid=mocked_liquid,
            volume=20,
        ),
        times=1,
    )


def test_diameter(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get the diameter from the core."""
    decoy.when(mock_well_core.diameter).then_return(12.3)

    assert subject.diameter == 12.3


def test_length(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get the length from the core."""
    decoy.when(mock_well_core.length).then_return(45.6)

    assert subject.length == 45.6


def test_width(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get the width from the core."""
    decoy.when(mock_well_core.width).then_return(78.9)

    assert subject.width == 78.9


def test_depth(decoy: Decoy, mock_well_core: WellCore, subject: Well) -> None:
    """It should get the depth from the core."""
    decoy.when(mock_well_core.depth).then_return(42.0)

    assert subject.depth == 42.0


def test_from_center_cartesian(
    decoy: Decoy, mock_well_core: WellCore, subject: Well
) -> None:
    """It should get the calculated center cartesian point from the core."""
    decoy.when(mock_well_core.from_center_cartesian(1, 2, 3)).then_return(
        Point(4, 5, 6)
    )

    result = subject.from_center_cartesian(1, 2, 3)

    assert result == Point(4, 5, 6)



"""Tests for opentrons.protocols.geometry.deck_conflict."""

from contextlib import nullcontext
from typing import ContextManager

import pytest

from opentrons_shared_data.labware.types import LabwareUri
from opentrons_shared_data.robot.types import RobotType

from opentrons.motion_planning import deck_conflict
from opentrons.types import DeckSlotName, StagingSlotName


@pytest.mark.parametrize(
    "robot_type, slot_name",
    [
        ("OT-2 Standard", DeckSlotName.SLOT_1),
        ("OT-3 Standard", DeckSlotName.SLOT_A1),
        ("OT-3 Standard", DeckSlotName.SLOT_A3),
    ],
)
def test_empty_no_conflict(robot_type: RobotType, slot_name: DeckSlotName) -> None:
    """It should not raise on empty input."""
    deck_conflict.check(
        existing_items={},
        new_item=deck_conflict.OtherModule(
            highest_z_including_labware=123, name_for_errors="foo"
        ),
        new_location=slot_name,
        robot_type=robot_type,
    )


@pytest.mark.parametrize(
    "robot_type, slot_name",
    [
        ("OT-2 Standard", DeckSlotName.SLOT_1),
        ("OT-3 Standard", DeckSlotName.SLOT_A1),
        ("OT-3 Standard", StagingSlotName.SLOT_A4),
    ],
)
def test_no_multiple_locations(robot_type: RobotType, slot_name: DeckSlotName) -> None:
    """It should not allow two items in the same slot."""
    item_1 = deck_conflict.Labware(
        uri=LabwareUri("some_labware_uri"),
        highest_z=123,
        is_fixed_trash=False,
        name_for_errors="some_item_1",
    )
    item_2 = deck_conflict.Labware(
        uri=LabwareUri("some_labware_uri"),
        highest_z=123,
        is_fixed_trash=False,
        name_for_errors="some_item_2",
    )

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=f"some_item_1 in slot {slot_name} prevents some_item_2 from using slot {slot_name}",
    ):
        deck_conflict.check(
            existing_items={slot_name: item_1},
            new_item=item_2,
            new_location=slot_name,
            robot_type=robot_type,
        )


@pytest.mark.parametrize(
    ("thermocycler_is_semi", "labware_location", "labware_should_be_allowed"),
    [
        # Non-semi config:
        (False, DeckSlotName.SLOT_1, True),
        (False, DeckSlotName.SLOT_7, False),
        (False, DeckSlotName.SLOT_8, False),
        (False, DeckSlotName.SLOT_10, False),
        (False, DeckSlotName.SLOT_11, False),
        # Semi config:
        (True, DeckSlotName.SLOT_1, True),
        (True, DeckSlotName.SLOT_7, False),
        (True, DeckSlotName.SLOT_8, True),
        (True, DeckSlotName.SLOT_10, False),
        (True, DeckSlotName.SLOT_11, True),
    ],
)
def test_labware_when_thermocycler(
    thermocycler_is_semi: bool,
    labware_location: DeckSlotName,
    labware_should_be_allowed: bool,
) -> None:
    """It should reject labware if a Thermocycler covers the same slot."""
    thermocycler = deck_conflict.ThermocyclerModule(
        name_for_errors="some_thermocycler",
        highest_z_including_labware=123,
        is_semi_configuration=thermocycler_is_semi,
    )

    labware = deck_conflict.Labware(
        uri=LabwareUri("some_labware_uri"),
        highest_z=123,
        is_fixed_trash=False,
        name_for_errors="some_labware",
    )

    maybe_raises: ContextManager[object]
    if labware_should_be_allowed:
        maybe_raises = nullcontext()  # Expecct no exception.
    else:
        maybe_raises = pytest.raises(  # Expect an exception..
            deck_conflict.DeckConflictError,
            match=(
                "some_thermocycler in slot 7 prevents"
                f" some_labware from using slot {labware_location}"
            ),
        )
    with maybe_raises:
        deck_conflict.check(
            existing_items={DeckSlotName.SLOT_7: thermocycler},
            new_location=labware_location,
            new_item=labware,
            robot_type="OT-2 Standard",
        )

    if labware_should_be_allowed:
        maybe_raises = nullcontext()  # Expecct no exception.
    else:
        maybe_raises = pytest.raises(  # Expect an exception..
            deck_conflict.DeckConflictError,
            match=(
                f"some_labware in slot {labware_location}"
                " prevents some_thermocycler from using slot 7"
            ),
        )
    with maybe_raises:
        deck_conflict.check(
            existing_items={labware_location: labware},
            new_location=DeckSlotName.SLOT_7,
            new_item=thermocycler,
            robot_type="OT-2 Standard",
        )


@pytest.mark.parametrize(
    ("labware_location", "labware_should_be_allowed"),
    [
        (DeckSlotName.SLOT_D1, True),
        (DeckSlotName.SLOT_B1, False),
        (DeckSlotName.SLOT_B2, True),
        (DeckSlotName.SLOT_A1, False),
        (DeckSlotName.SLOT_A2, True),
    ],
)
def test_flex_labware_when_thermocycler(
    labware_location: DeckSlotName,
    labware_should_be_allowed: bool,
) -> None:
    """It should reject labware if a Thermocycler covers the same slot."""
    thermocycler = deck_conflict.ThermocyclerModule(
        name_for_errors="some_thermocycler",
        highest_z_including_labware=123,
        is_semi_configuration=False,
    )

    labware = deck_conflict.Labware(
        uri=LabwareUri("some_labware_uri"),
        highest_z=123,
        is_fixed_trash=False,
        name_for_errors="some_labware",
    )

    maybe_raises: ContextManager[object]
    if labware_should_be_allowed:
        maybe_raises = nullcontext()  # Expecct no exception.
    else:
        maybe_raises = pytest.raises(  # Expect an exception..
            deck_conflict.DeckConflictError,
            match=(
                "some_thermocycler in slot B1 prevents"
                f" some_labware from using slot {labware_location}"
            ),
        )
    with maybe_raises:
        deck_conflict.check(
            existing_items={DeckSlotName.SLOT_B1: thermocycler},
            new_location=labware_location,
            new_item=labware,
            robot_type="OT-3 Standard",
        )

    if labware_should_be_allowed:
        maybe_raises = nullcontext()  # Expecct no exception.
    else:
        maybe_raises = pytest.raises(  # Expect an exception..
            deck_conflict.DeckConflictError,
            match=(
                f"some_labware in slot {labware_location}"
                " prevents some_thermocycler from using slot B1"
            ),
        )
    with maybe_raises:
        deck_conflict.check(
            existing_items={labware_location: labware},
            new_location=DeckSlotName.SLOT_B1,
            new_item=thermocycler,
            robot_type="OT-3 Standard",
        )


def test_flex_trash_bin_blocks_thermocycler() -> None:
    """It should prevent loading a thermocycler when there is a trash in A1 and vice-versa."""
    thermocycler = deck_conflict.ThermocyclerModule(
        name_for_errors="some_thermocycler",
        highest_z_including_labware=123,
        is_semi_configuration=False,
    )
    trash = deck_conflict.TrashBin(name_for_errors="some_trash_bin", highest_z=1.23)

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            "some_trash_bin in slot A1 prevents some_thermocycler from using slot B1"
        ),
    ):
        deck_conflict.check(
            existing_items={DeckSlotName.SLOT_A1: trash},
            new_item=thermocycler,
            new_location=DeckSlotName.SLOT_B1,
            robot_type="OT-3 Standard",
        )
    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            "some_thermocycler in slot B1 prevents some_trash_bin from using slot A1"
        ),
    ):
        deck_conflict.check(
            existing_items={DeckSlotName.SLOT_B1: thermocycler},
            new_item=trash,
            new_location=DeckSlotName.SLOT_A1,
            robot_type="OT-3 Standard",
        )


@pytest.mark.parametrize(
    ("heater_shaker_location", "labware_location"),
    [
        (DeckSlotName.SLOT_1, DeckSlotName.SLOT_2),
        (DeckSlotName.SLOT_2, DeckSlotName.SLOT_1),
        (DeckSlotName.SLOT_2, DeckSlotName.SLOT_3),
        (DeckSlotName.SLOT_3, DeckSlotName.SLOT_2),
        (DeckSlotName.SLOT_4, DeckSlotName.SLOT_5),
        (DeckSlotName.SLOT_5, DeckSlotName.SLOT_4),
        (DeckSlotName.SLOT_5, DeckSlotName.SLOT_6),
        (DeckSlotName.SLOT_6, DeckSlotName.SLOT_5),
        (DeckSlotName.SLOT_7, DeckSlotName.SLOT_8),
        (DeckSlotName.SLOT_8, DeckSlotName.SLOT_7),
        (DeckSlotName.SLOT_8, DeckSlotName.SLOT_9),
        (DeckSlotName.SLOT_9, DeckSlotName.SLOT_8),
        (DeckSlotName.SLOT_10, DeckSlotName.SLOT_11),
        (DeckSlotName.SLOT_11, DeckSlotName.SLOT_10),
    ],
)
def test_labware_when_heater_shaker(
    heater_shaker_location: DeckSlotName,
    labware_location: DeckSlotName,
) -> None:
    """It should allow short labware east and west if a heater-shaker is placed."""
    heater_shaker = deck_conflict.HeaterShakerModule(
        highest_z_including_labware=123, name_for_errors="some_heater_shaker"
    )
    cool_labware = deck_conflict.Labware(
        uri=LabwareUri("test/cool_labware/1"),
        highest_z=1,
        is_fixed_trash=False,
        name_for_errors="cool_labware",
    )
    lame_labware = deck_conflict.Labware(
        uri=LabwareUri("test/lame_labware/1"),
        highest_z=999,
        is_fixed_trash=False,
        name_for_errors="lame_labware",
    )

    deck_conflict.check(
        existing_items={heater_shaker_location: heater_shaker},
        new_location=labware_location,
        new_item=cool_labware,
        robot_type="OT-2 Standard",
    )
    deck_conflict.check(
        existing_items={labware_location: cool_labware},
        new_location=heater_shaker_location,
        new_item=heater_shaker,
        robot_type="OT-2 Standard",
    )

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            f"some_heater_shaker in slot {heater_shaker_location}"
            f" prevents lame_labware from using slot {labware_location}"
        ),
    ):
        deck_conflict.check(
            existing_items={heater_shaker_location: heater_shaker},
            new_location=labware_location,
            new_item=lame_labware,
            robot_type="OT-2 Standard",
        )

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            f"lame_labware in slot {labware_location}"
            f" prevents some_heater_shaker from using slot {heater_shaker_location}"
        ),
    ):
        deck_conflict.check(
            existing_items={labware_location: lame_labware},
            new_location=heater_shaker_location,
            new_item=heater_shaker,
            robot_type="OT-2 Standard",
        )


@pytest.mark.parametrize(
    ("heater_shaker_location", "other_module_location"),
    [
        (DeckSlotName.SLOT_1, DeckSlotName.SLOT_2),
        (DeckSlotName.SLOT_1, DeckSlotName.SLOT_4),
        (DeckSlotName.SLOT_2, DeckSlotName.SLOT_1),
        (DeckSlotName.SLOT_2, DeckSlotName.SLOT_3),
        (DeckSlotName.SLOT_2, DeckSlotName.SLOT_5),
        (DeckSlotName.SLOT_3, DeckSlotName.SLOT_2),
        (DeckSlotName.SLOT_3, DeckSlotName.SLOT_6),
        (DeckSlotName.SLOT_4, DeckSlotName.SLOT_1),
        (DeckSlotName.SLOT_4, DeckSlotName.SLOT_5),
        (DeckSlotName.SLOT_4, DeckSlotName.SLOT_7),
        (DeckSlotName.SLOT_5, DeckSlotName.SLOT_2),
        (DeckSlotName.SLOT_5, DeckSlotName.SLOT_4),
        (DeckSlotName.SLOT_5, DeckSlotName.SLOT_6),
        (DeckSlotName.SLOT_5, DeckSlotName.SLOT_8),
        (DeckSlotName.SLOT_6, DeckSlotName.SLOT_3),
        (DeckSlotName.SLOT_6, DeckSlotName.SLOT_5),
        (DeckSlotName.SLOT_6, DeckSlotName.SLOT_9),
        (DeckSlotName.SLOT_7, DeckSlotName.SLOT_4),
        (DeckSlotName.SLOT_7, DeckSlotName.SLOT_8),
        (DeckSlotName.SLOT_7, DeckSlotName.SLOT_10),
        (DeckSlotName.SLOT_8, DeckSlotName.SLOT_5),
        (DeckSlotName.SLOT_8, DeckSlotName.SLOT_7),
        (DeckSlotName.SLOT_8, DeckSlotName.SLOT_9),
        (DeckSlotName.SLOT_8, DeckSlotName.SLOT_11),
        (DeckSlotName.SLOT_9, DeckSlotName.SLOT_6),
        (DeckSlotName.SLOT_9, DeckSlotName.SLOT_8),
        (DeckSlotName.SLOT_10, DeckSlotName.SLOT_7),
        (DeckSlotName.SLOT_10, DeckSlotName.SLOT_11),
        (DeckSlotName.SLOT_11, DeckSlotName.SLOT_8),
        (DeckSlotName.SLOT_11, DeckSlotName.SLOT_10),
    ],
)
def test_no_modules_when_heater_shaker(
    heater_shaker_location: DeckSlotName,
    other_module_location: DeckSlotName,
) -> None:
    """It should not allow other modules north and south of the H/S.

    All modules are taller than the H/S height restriction,
    so this test only checks modules specifically in N/S.
    """
    heater_shaker = deck_conflict.HeaterShakerModule(
        highest_z_including_labware=123, name_for_errors="some_heater_shaker"
    )
    other_module = deck_conflict.OtherModule(
        highest_z_including_labware=0, name_for_errors="some_other_module"
    )

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            f"some_heater_shaker in slot {heater_shaker_location}"
            f" prevents some_other_module from using slot {other_module_location}"
        ),
    ):
        deck_conflict.check(
            existing_items={heater_shaker_location: heater_shaker},
            new_location=other_module_location,
            new_item=other_module,
            robot_type="OT-2 Standard",
        )

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            f"some_other_module in slot {other_module_location}"
            f" prevents some_heater_shaker from using slot {heater_shaker_location}"
        ),
    ):
        deck_conflict.check(
            existing_items={other_module_location: other_module},
            new_location=heater_shaker_location,
            new_item=heater_shaker,
            robot_type="OT-2 Standard",
        )


@pytest.mark.parametrize(
    "allowed_tip_rack_load_name",
    deck_conflict.HS_ALLOWED_ADJACENT_TALL_LABWARE,
)
@pytest.mark.parametrize("allowed_tip_rack_version", [1, 2])
@pytest.mark.parametrize(
    ("heater_shaker_location", "tip_rack_location"),
    [
        (DeckSlotName.SLOT_1, DeckSlotName.SLOT_2),
        (DeckSlotName.SLOT_2, DeckSlotName.SLOT_1),
        (DeckSlotName.SLOT_2, DeckSlotName.SLOT_3),
        (DeckSlotName.SLOT_3, DeckSlotName.SLOT_2),
        (DeckSlotName.SLOT_4, DeckSlotName.SLOT_5),
        (DeckSlotName.SLOT_5, DeckSlotName.SLOT_4),
        (DeckSlotName.SLOT_5, DeckSlotName.SLOT_6),
        (DeckSlotName.SLOT_6, DeckSlotName.SLOT_5),
        (DeckSlotName.SLOT_7, DeckSlotName.SLOT_8),
        (DeckSlotName.SLOT_8, DeckSlotName.SLOT_7),
        (DeckSlotName.SLOT_8, DeckSlotName.SLOT_9),
        (DeckSlotName.SLOT_9, DeckSlotName.SLOT_8),
        (DeckSlotName.SLOT_10, DeckSlotName.SLOT_11),
        (DeckSlotName.SLOT_11, DeckSlotName.SLOT_10),
    ],
)
def test_tip_rack_when_heater_shaker(
    allowed_tip_rack_load_name: str,
    allowed_tip_rack_version: int,
    heater_shaker_location: DeckSlotName,
    tip_rack_location: DeckSlotName,
) -> None:
    """It should allow short tip racks east and west if a heater-shaker is placed."""
    heater_shaker = deck_conflict.HeaterShakerModule(
        highest_z_including_labware=123,
        name_for_errors="some_heater_shaker",
    )

    too_high = deck_conflict.HS_MAX_X_ADJACENT_ITEM_HEIGHT + 0.1

    cool_tip_rack_uri = LabwareUri(
        f"opentrons/{allowed_tip_rack_load_name}/{allowed_tip_rack_version}"
    )
    cool_tip_rack = deck_conflict.Labware(
        uri=cool_tip_rack_uri,
        highest_z=too_high,
        is_fixed_trash=False,
        name_for_errors="cool_tip_rack",
    )

    lame_tip_rack = deck_conflict.Labware(
        uri=LabwareUri("test/lame_tip_rack/1"),
        highest_z=too_high,
        is_fixed_trash=False,
        name_for_errors="lame_tip_rack",
    )

    deck_conflict.check(
        existing_items={heater_shaker_location: heater_shaker},
        new_location=tip_rack_location,
        new_item=cool_tip_rack,
        robot_type="OT-2 Standard",
    )
    deck_conflict.check(
        existing_items={tip_rack_location: cool_tip_rack},
        new_location=heater_shaker_location,
        new_item=heater_shaker,
        robot_type="OT-2 Standard",
    )

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            f"some_heater_shaker in slot {heater_shaker_location}"
            f" prevents lame_tip_rack from using slot {tip_rack_location}"
        ),
    ):
        deck_conflict.check(
            existing_items={heater_shaker_location: heater_shaker},
            new_location=tip_rack_location,
            new_item=lame_tip_rack,
            robot_type="OT-2 Standard",
        )

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            f"lame_tip_rack in slot {tip_rack_location}"
            f" prevents some_heater_shaker from using slot {heater_shaker_location}"
        ),
    ):
        deck_conflict.check(
            existing_items={tip_rack_location: lame_tip_rack},
            new_location=heater_shaker_location,
            new_item=heater_shaker,
            robot_type="OT-2 Standard",
        )


def test_no_heater_shaker_south_of_trash() -> None:
    """It should prevent loading a Heater-Shaker south of a fixed trash.

    This is because the trash prevents accessing the Heater-Shaker's anchor screw.

    Unrelated restrictions on height may prevent putting a Heater-Shaker in other places
    around the fixed trash. Those restrictions should be covered by HS<->labware tests.
    """
    heater_shaker = deck_conflict.HeaterShakerModule(
        highest_z_including_labware=123, name_for_errors="some_heater_shaker"
    )
    trash = deck_conflict.Labware(
        # We want to test that the subject rejects the HS placement on account of this
        # labware being a fixed trash, not on account of this labware being too tall.
        highest_z=0,
        name_for_errors="some_fixed_trash",
        uri=LabwareUri("test/some_fixed_trash/1"),
        is_fixed_trash=True,
    )

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            "some_fixed_trash in slot 12 prevents some_heater_shaker from using slot 9"
        ),
    ):
        deck_conflict.check(
            existing_items={DeckSlotName.FIXED_TRASH: trash},
            new_item=heater_shaker,
            new_location=DeckSlotName.SLOT_9,
            robot_type="OT-2 Standard",
        )


def test_heater_shaker_restrictions_trash_bin_addressable_area() -> None:
    """It should prevent loading a Heater-Shaker adjacent of a non-labware trash bin.

    This is for the OT-2 only and for slot 11 and slot 9
    """
    heater_shaker = deck_conflict.HeaterShakerModule(
        highest_z_including_labware=123, name_for_errors="some_heater_shaker"
    )
    trash = deck_conflict.TrashBin(name_for_errors="some_trash_bin", highest_z=456)

    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            "some_trash_bin in slot 12 prevents some_heater_shaker from using slot 9"
        ),
    ):
        deck_conflict.check(
            existing_items={DeckSlotName.FIXED_TRASH: trash},
            new_item=heater_shaker,
            new_location=DeckSlotName.SLOT_9,
            robot_type="OT-2 Standard",
        )
    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            "some_trash_bin in slot 12 prevents some_heater_shaker from using slot 11"
        ),
    ):
        deck_conflict.check(
            existing_items={DeckSlotName.FIXED_TRASH: trash},
            new_item=heater_shaker,
            new_location=DeckSlotName.SLOT_11,
            robot_type="OT-2 Standard",
        )


@pytest.mark.parametrize(
    ("deck_slot_name", "adjacent_staging_slot", "non_adjacent_staging_slot"),
    [
        (DeckSlotName.SLOT_A3, StagingSlotName.SLOT_A4, StagingSlotName.SLOT_B4),
        (DeckSlotName.SLOT_B3, StagingSlotName.SLOT_B4, StagingSlotName.SLOT_C4),
        (DeckSlotName.SLOT_C3, StagingSlotName.SLOT_C4, StagingSlotName.SLOT_D4),
        (DeckSlotName.SLOT_D3, StagingSlotName.SLOT_D4, StagingSlotName.SLOT_A4),
    ],
)
def test_no_staging_slot_adjacent_to_module(
    deck_slot_name: DeckSlotName,
    adjacent_staging_slot: StagingSlotName,
    non_adjacent_staging_slot: StagingSlotName,
) -> None:
    """It should raise if certain modules are placed adjacent to labware on a staging slot."""
    staging_slot_labware = deck_conflict.Labware(
        uri=LabwareUri("some_labware_uri"),
        highest_z=123,
        is_fixed_trash=False,
        name_for_errors="some_labware",
    )
    heater_shaker = deck_conflict.HeaterShakerModule(
        name_for_errors="some_heater_shaker",
        highest_z_including_labware=123,
    )
    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            f"some_labware in slot {adjacent_staging_slot}"
            f" prevents some_heater_shaker from using slot {deck_slot_name}"
        ),
    ):
        deck_conflict.check(
            existing_items={adjacent_staging_slot: staging_slot_labware},
            new_item=heater_shaker,
            new_location=deck_slot_name,
            robot_type="OT-3 Standard",
        )

    # Non-adjacent staging slot passes
    deck_conflict.check(
        existing_items={non_adjacent_staging_slot: staging_slot_labware},
        new_item=heater_shaker,
        new_location=deck_slot_name,
        robot_type="OT-3 Standard",
    )

    other_module = deck_conflict.OtherModule(
        name_for_errors="some_other_module",
        highest_z_including_labware=123,
    )
    with pytest.raises(
        deck_conflict.DeckConflictError,
        match=(
            f"some_other_module in slot {deck_slot_name}"
            f" prevents some_labware from using slot {adjacent_staging_slot}"
        ),
    ):
        deck_conflict.check(
            existing_items={deck_slot_name: other_module},
            new_item=staging_slot_labware,
            new_location=adjacent_staging_slot,
            robot_type="OT-3 Standard",
        )

    # Magnetic block is allowed
    magnetic_block = deck_conflict.MagneticBlockModule(
        name_for_errors="some_mag_block",
        highest_z_including_labware=123,
    )
    deck_conflict.check(
        existing_items={adjacent_staging_slot: staging_slot_labware},
        new_item=magnetic_block,
        new_location=deck_slot_name,
        robot_type="OT-3 Standard",
    )


@pytest.mark.parametrize(
    ("deck_slot_name",),
    [
        (DeckSlotName.SLOT_A3,),
        (DeckSlotName.SLOT_B3,),
        (DeckSlotName.SLOT_C3,),
        (DeckSlotName.SLOT_D3,),
    ],
)
def test_allow_labware_in_stacker_quote_slot_unquote(
    deck_slot_name: DeckSlotName,
) -> None:
    """You should be able to load labware "in" "the same" "slot" as a stacker."""
    labware = deck_conflict.Labware(
        uri=LabwareUri("some_labware_uri"),
        highest_z=123,
        is_fixed_trash=False,
        name_for_errors="some_labware",
    )
    stacker = deck_conflict.FlexStackerModule(
        name_for_errors="some_stacker",
        highest_z_including_labware=123,
    )
    deck_conflict.check(
        existing_items={deck_slot_name: stacker},
        new_item=labware,
        new_location=deck_slot_name,
        robot_type="OT-3 Standard",
    )
    deck_conflict.check(
        existing_items={deck_slot_name: labware},
        new_item=stacker,
        new_location=deck_slot_name,
        robot_type="OT-3 Standard",
    )



from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Tuple, Union

from numpy import interp
from pydantic import ValidationError

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    AspirateProperties as SharedDataAspirateProperties,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    BlowoutLocation,
    Coordinate,
    PositionReference,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    BlowoutParams as SharedDataBlowoutParams,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    BlowoutProperties as SharedDataBlowoutProperties,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    ByTipTypeSetting as SharedByTipTypeSetting,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    DelayParams as SharedDataDelayParams,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    DelayProperties as SharedDataDelayProperties,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassTouchTipParams as SharedDataTouchTipParams,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    MixParams as SharedDataMixParams,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    MixProperties as SharedDataMixProperties,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    MultiDispenseProperties as SharedDataMultiDispenseProperties,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    RetractAspirate as SharedDataRetractAspirate,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    RetractDispense as SharedDataRetractDispense,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    SingleDispenseProperties as SharedDataSingleDispenseProperties,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    Submerge as SharedDataSubmerge,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    TipPosition as SharedDataTipPosition,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    TouchTipProperties as SharedDataTouchTipProperties,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    TransferProperties as SharedDataTransferProperties,
)
from opentrons_shared_data.liquid_classes.types import TipPositionDict

from . import validation


class LiquidHandlingPropertyByVolume:
    def __init__(self, by_volume_property: Sequence[Tuple[float, float]]) -> None:
        self._initial_properties_by_volume = by_volume_property
        self._properties_by_volume: Dict[float, float] = {
            float(volume): value for volume, value in by_volume_property
        }
        # Volumes need to be sorted for proper interpolation of non-defined volumes, and the
        # corresponding values need to be in the same order for them to be interpolated correctly
        self._sorted_volumes: Tuple[float, ...] = ()
        self._sorted_values: Tuple[float, ...] = ()
        self._sort_volume_and_values()

    def as_dict(self) -> Dict[float, float]:
        """Get a dictionary representation of all set volumes and values along with the default."""
        return self._properties_by_volume

    def as_list_of_tuples(self) -> List[Tuple[float, float]]:
        """Get as list of tuples."""
        return list(self._properties_by_volume.items())

    def get_for_volume(self, volume: float) -> float:
        """Get a value by volume for this property. Volumes not defined will be interpolated between set volumes."""
        validated_volume = validation.ensure_positive_float(volume)
        if len(self._properties_by_volume) == 0:
            raise ValueError(
                "No properties found for any volumes. Cannot interpolate for the given volume."
            )
        try:
            return self._properties_by_volume[validated_volume]
        except KeyError:
            # If volume is not defined in dictionary, do a piecewise interpolation with existing sorted values
            return float(
                interp(validated_volume, self._sorted_volumes, self._sorted_values)
            )

    def set_for_all_volumes(self, value: float) -> None:
        """Override all existing volume-dependent values with the given value."""
        self.clear_values()
        self.set_for_volume(0, value)

    def set_for_volume(self, volume: float, value: float) -> None:
        """Add a new volume and value for the property for the interpolation curve."""
        validated_volume = validation.ensure_positive_float(volume)
        self._properties_by_volume[validated_volume] = value
        self._sort_volume_and_values()

    def delete_for_volume(self, volume: float) -> None:
        """Remove an existing volume and value from the property."""
        try:
            del self._properties_by_volume[volume]
        except KeyError:
            raise KeyError(f"No value set for volume {volume} uL")
        self._sort_volume_and_values()

    def clear_values(self) -> None:
        """Removes all existing volume and value pairs from the curve."""
        self._properties_by_volume = {}

    def reset_values(self) -> None:
        """Resets volumes and values to the default."""
        self._properties_by_volume = {
            float(volume): value for volume, value in self._initial_properties_by_volume
        }
        self._sort_volume_and_values()

    def _sort_volume_and_values(self) -> None:
        """Sort volume in increasing order along with corresponding values in matching order."""
        self._sorted_volumes, self._sorted_values = (
            zip(*sorted(self._properties_by_volume.items()))
            if len(self._properties_by_volume) > 0
            else [(), ()]
        )


# We use slots for this dataclass (and the rest of liquid properties) to prevent dynamic creation of attributes
# not defined in the class, not for any performance reasons. This is so that mistyping properties when overriding
# values will cause the protocol to fail analysis, rather than silently passing.
@dataclass(slots=True)
class TipPosition:
    _position_reference: PositionReference
    _offset: Coordinate

    @property
    def position_reference(self) -> PositionReference:
        return self._position_reference

    @position_reference.setter
    def position_reference(self, new_position: Union[str, PositionReference]) -> None:
        self._position_reference = (
            new_position
            if isinstance(new_position, PositionReference)
            else PositionReference(new_position)
        )

    @property
    def offset(self) -> Coordinate:
        return self._offset

    @offset.setter
    def offset(self, new_offset: Union[Sequence[float], Coordinate]) -> None:
        if isinstance(new_offset, Coordinate):
            new_coordinate: Sequence[Union[int, float]] = [
                new_offset.x,
                new_offset.y,
                new_offset.z,
            ]
        else:
            new_coordinate = new_offset
        x, y, z = validation.validate_coordinates(new_coordinate)
        self._offset = Coordinate(x=x, y=y, z=z)

    def as_shared_data_model(self) -> SharedDataTipPosition:
        return SharedDataTipPosition(
            positionReference=self._position_reference,
            offset=self.offset,
        )


@dataclass(slots=True)
class DelayProperties:
    _enabled: bool
    _duration: Optional[float]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enable: bool) -> None:
        validated_enable = validation.ensure_boolean(enable)
        if validated_enable and self._duration is None:
            raise ValueError("duration must be set before enabling delay.")
        self._enabled = validated_enable

    @property
    def duration(self) -> Optional[float]:
        return self._duration

    @duration.setter
    def duration(self, new_duration: float) -> None:
        validated_duration = validation.ensure_positive_float(new_duration)
        self._duration = validated_duration

    def as_shared_data_model(self) -> SharedDataDelayProperties:
        return SharedDataDelayProperties(
            enable=self._enabled,
            params=SharedDataDelayParams(duration=self.duration)
            if self.duration is not None
            else None,
        )


@dataclass(slots=True)
class TouchTipProperties:
    _enabled: bool
    _z_offset: Optional[float]
    _mm_from_edge: Optional[float]
    _speed: Optional[float]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enable: bool) -> None:
        validated_enable = validation.ensure_boolean(enable)
        if validated_enable and (
            self._z_offset is None or self._mm_from_edge is None or self._speed is None
        ):
            raise ValueError(
                "z_offset, mm_from_edge and speed must be set before enabling touch tip."
            )
        self._enabled = validated_enable

    @property
    def z_offset(self) -> Optional[float]:
        return self._z_offset

    @z_offset.setter
    def z_offset(self, new_offset: float) -> None:
        validated_offset = validation.ensure_float(new_offset)
        self._z_offset = validated_offset

    @property
    def mm_from_edge(self) -> Optional[float]:
        return self._mm_from_edge

    @mm_from_edge.setter
    def mm_from_edge(self, new_mm: float) -> None:
        validated_mm = validation.ensure_float(new_mm)
        self._mm_from_edge = validated_mm

    @property
    def speed(self) -> Optional[float]:
        return self._speed

    @speed.setter
    def speed(self, new_speed: float) -> None:
        validated_speed = validation.ensure_greater_than_zero_float(new_speed)
        self._speed = validated_speed

    def _get_shared_data_params(self) -> Optional[SharedDataTouchTipParams]:
        """Get the touch tip params in schema v1 shape."""
        if (
            self._z_offset is not None
            and self._mm_from_edge is not None
            and self._speed is not None
        ):
            return SharedDataTouchTipParams(
                zOffset=self._z_offset,
                mmFromEdge=self._mm_from_edge,
                speed=self._speed,
            )
        else:
            return None

    def as_shared_data_model(self) -> SharedDataTouchTipProperties:
        return SharedDataTouchTipProperties(
            enable=self._enabled,
            params=self._get_shared_data_params(),
        )


@dataclass(slots=True)
class MixProperties:
    _enabled: bool
    _repetitions: Optional[int]
    _volume: Optional[float]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enable: bool) -> None:
        validated_enable = validation.ensure_boolean(enable)
        if validated_enable and (self._repetitions is None or self._volume is None):
            raise ValueError("repetitions and volume must be set before enabling mix.")
        self._enabled = validated_enable

    @property
    def repetitions(self) -> Optional[int]:
        return self._repetitions

    @repetitions.setter
    def repetitions(self, new_repetitions: int) -> None:
        validated_repetitions = validation.ensure_positive_int(new_repetitions)
        self._repetitions = validated_repetitions

    @property
    def volume(self) -> Optional[float]:
        return self._volume

    @volume.setter
    def volume(self, new_volume: float) -> None:
        validated_volume = validation.ensure_greater_than_zero_float(new_volume)
        self._volume = validated_volume

    def _get_shared_data_params(self) -> Optional[SharedDataMixParams]:
        """Get the mix params in schema v1 shape."""
        if self._repetitions is not None and self._volume is not None:
            return SharedDataMixParams(
                repetitions=self._repetitions,
                volume=self._volume,
            )
        else:
            return None

    def as_shared_data_model(self) -> SharedDataMixProperties:
        return SharedDataMixProperties(
            enable=self._enabled,
            params=self._get_shared_data_params(),
        )


@dataclass(slots=True)
class BlowoutProperties:
    _enabled: bool
    _location: Optional[BlowoutLocation]
    _flow_rate: Optional[float]
    _blowout_position: Optional[TipPosition]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enable: bool) -> None:
        validated_enable = validation.ensure_boolean(enable)
        if validated_enable and (self._location is None or self._flow_rate is None):
            raise ValueError(
                "location and flow_rate must be set before enabling blowout."
            )
        self._enabled = validated_enable

    @property
    def location(self) -> Optional[BlowoutLocation]:
        return self._location

    @location.setter
    def location(self, new_location: str) -> None:
        self._location = BlowoutLocation(new_location)

    @property
    def flow_rate(self) -> Optional[float]:
        return self._flow_rate

    @flow_rate.setter
    def flow_rate(self, new_flow_rate: float) -> None:
        validated_flow_rate = validation.ensure_greater_than_zero_float(new_flow_rate)
        self._flow_rate = validated_flow_rate

    @property
    def blowout_position(self) -> Optional[TipPosition]:
        return self._blowout_position

    @blowout_position.setter
    def blowout_position(
        self, new_position: Union[TipPosition, TipPositionDict, None]
    ) -> None:
        if new_position is None:
            self._blowout_position = None
        else:
            self._blowout_position = _ensure_validated_tip_position(new_position)

    def _get_shared_data_params(self) -> Optional[SharedDataBlowoutParams]:
        """Get the mix params in schema v1 shape."""
        if self._location is not None and self._flow_rate is not None:
            return SharedDataBlowoutParams(
                location=self._location,
                flowRate=self._flow_rate,
                blowoutPosition=self._blowout_position.as_shared_data_model()
                if self._blowout_position is not None
                else None,
            )
        else:
            return None

    def as_shared_data_model(self) -> SharedDataBlowoutProperties:
        return SharedDataBlowoutProperties(
            enable=self._enabled,
            params=self._get_shared_data_params(),
        )


@dataclass(slots=True)
class _SubmergeRetractCommon:
    _speed: float
    _delay: DelayProperties

    @property
    def speed(self) -> float:
        return self._speed

    @speed.setter
    def speed(self, new_speed: float) -> None:
        validated_speed = validation.ensure_positive_float(new_speed)
        self._speed = validated_speed

    @property
    def delay(self) -> DelayProperties:
        return self._delay


@dataclass(slots=True)
class Submerge(_SubmergeRetractCommon):
    _start_position: TipPosition

    @property
    def start_position(self) -> TipPosition:
        return self._start_position

    @start_position.setter
    def start_position(self, new_position: TipPosition) -> None:
        self._start_position = _ensure_validated_tip_position(new_position)

    def as_shared_data_model(self) -> SharedDataSubmerge:
        return SharedDataSubmerge(
            startPosition=self._start_position.as_shared_data_model(),
            speed=self._speed,
            delay=self._delay.as_shared_data_model(),
        )


@dataclass(slots=True)
class RetractAspirate(_SubmergeRetractCommon):
    _end_position: TipPosition
    _air_gap_by_volume: LiquidHandlingPropertyByVolume
    _touch_tip: TouchTipProperties

    @property
    def end_position(self) -> TipPosition:
        return self._end_position

    @end_position.setter
    def end_position(self, new_position: TipPosition) -> None:
        self._end_position = _ensure_validated_tip_position(new_position)

    @property
    def air_gap_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._air_gap_by_volume

    @property
    def touch_tip(self) -> TouchTipProperties:
        return self._touch_tip

    def as_shared_data_model(self) -> SharedDataRetractAspirate:
        return SharedDataRetractAspirate(
            endPosition=self._end_position.as_shared_data_model(),
            speed=self._speed,
            airGapByVolume=self._air_gap_by_volume.as_list_of_tuples(),
            touchTip=self._touch_tip.as_shared_data_model(),
            delay=self._delay.as_shared_data_model(),
        )


@dataclass(slots=True)
class RetractDispense(_SubmergeRetractCommon):
    _end_position: TipPosition
    _air_gap_by_volume: LiquidHandlingPropertyByVolume
    _touch_tip: TouchTipProperties
    _blowout: BlowoutProperties

    @property
    def end_position(self) -> TipPosition:
        return self._end_position

    @end_position.setter
    def end_position(self, new_position: TipPosition) -> None:
        self._end_position = _ensure_validated_tip_position(new_position)

    @property
    def air_gap_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._air_gap_by_volume

    @property
    def touch_tip(self) -> TouchTipProperties:
        return self._touch_tip

    @property
    def blowout(self) -> BlowoutProperties:
        return self._blowout

    def as_shared_data_model(self) -> SharedDataRetractDispense:
        return SharedDataRetractDispense(
            endPosition=self._end_position.as_shared_data_model(),
            speed=self._speed,
            airGapByVolume=self._air_gap_by_volume.as_list_of_tuples(),
            blowout=self._blowout.as_shared_data_model(),
            touchTip=self._touch_tip.as_shared_data_model(),
            delay=self._delay.as_shared_data_model(),
        )


@dataclass(slots=True)
class _BaseLiquidHandlingProperties:
    _submerge: Submerge
    _flow_rate_by_volume: LiquidHandlingPropertyByVolume
    _correction_by_volume: LiquidHandlingPropertyByVolume
    _delay: DelayProperties

    @property
    def submerge(self) -> Submerge:
        return self._submerge

    @property
    def flow_rate_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._flow_rate_by_volume

    @property
    def correction_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._correction_by_volume

    @property
    def delay(self) -> DelayProperties:
        return self._delay


@dataclass(slots=True)
class AspirateProperties(_BaseLiquidHandlingProperties):
    _aspirate_position: TipPosition
    _retract: RetractAspirate
    _pre_wet: bool
    _mix: MixProperties

    @property
    def aspirate_position(self) -> TipPosition:
        return self._aspirate_position

    @aspirate_position.setter
    def aspirate_position(self, new_position: TipPosition) -> None:
        self._aspirate_position = _ensure_validated_tip_position(new_position)

    @property
    def pre_wet(self) -> bool:
        return self._pre_wet

    @pre_wet.setter
    def pre_wet(self, new_setting: bool) -> None:
        validated_setting = validation.ensure_boolean(new_setting)
        self._pre_wet = validated_setting

    @property
    def retract(self) -> RetractAspirate:
        return self._retract

    @property
    def mix(self) -> MixProperties:
        return self._mix

    def as_shared_data_model(self) -> SharedDataAspirateProperties:
        return SharedDataAspirateProperties(
            submerge=self._submerge.as_shared_data_model(),
            retract=self._retract.as_shared_data_model(),
            aspiratePosition=self._aspirate_position.as_shared_data_model(),
            flowRateByVolume=self._flow_rate_by_volume.as_list_of_tuples(),
            preWet=self._pre_wet,
            mix=self._mix.as_shared_data_model(),
            delay=self._delay.as_shared_data_model(),
            correctionByVolume=self._correction_by_volume.as_list_of_tuples(),
        )


@dataclass(slots=True)
class SingleDispenseProperties(_BaseLiquidHandlingProperties):
    _dispense_position: TipPosition
    _retract: RetractDispense
    _push_out_by_volume: LiquidHandlingPropertyByVolume
    _mix: MixProperties

    @property
    def dispense_position(self) -> TipPosition:
        return self._dispense_position

    @dispense_position.setter
    def dispense_position(self, new_position: TipPosition) -> None:
        self._dispense_position = _ensure_validated_tip_position(new_position)

    @property
    def push_out_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._push_out_by_volume

    @property
    def retract(self) -> RetractDispense:
        return self._retract

    @property
    def mix(self) -> MixProperties:
        return self._mix

    def as_shared_data_model(self) -> SharedDataSingleDispenseProperties:
        return SharedDataSingleDispenseProperties(
            submerge=self._submerge.as_shared_data_model(),
            retract=self._retract.as_shared_data_model(),
            dispensePosition=self._dispense_position.as_shared_data_model(),
            flowRateByVolume=self._flow_rate_by_volume.as_list_of_tuples(),
            mix=self._mix.as_shared_data_model(),
            pushOutByVolume=self._push_out_by_volume.as_list_of_tuples(),
            delay=self._delay.as_shared_data_model(),
            correctionByVolume=self._correction_by_volume.as_list_of_tuples(),
        )


@dataclass(slots=True)
class MultiDispenseProperties(_BaseLiquidHandlingProperties):
    _dispense_position: TipPosition
    _retract: RetractDispense
    _conditioning_by_volume: LiquidHandlingPropertyByVolume
    _disposal_by_volume: LiquidHandlingPropertyByVolume

    @property
    def dispense_position(self) -> TipPosition:
        return self._dispense_position

    @dispense_position.setter
    def dispense_position(self, new_position: TipPosition) -> None:
        self._dispense_position = _ensure_validated_tip_position(new_position)

    @property
    def retract(self) -> RetractDispense:
        return self._retract

    @property
    def conditioning_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._conditioning_by_volume

    @property
    def disposal_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._disposal_by_volume

    def as_shared_data_model(self) -> SharedDataMultiDispenseProperties:
        return SharedDataMultiDispenseProperties(
            submerge=self._submerge.as_shared_data_model(),
            retract=self._retract.as_shared_data_model(),
            dispensePosition=self._dispense_position.as_shared_data_model(),
            flowRateByVolume=self._flow_rate_by_volume.as_list_of_tuples(),
            conditioningByVolume=self._conditioning_by_volume.as_list_of_tuples(),
            disposalByVolume=self._disposal_by_volume.as_list_of_tuples(),
            delay=self._delay.as_shared_data_model(),
            correctionByVolume=self._correction_by_volume.as_list_of_tuples(),
        )


@dataclass(slots=True)
class TransferProperties:
    _aspirate: AspirateProperties
    _dispense: SingleDispenseProperties
    _multi_dispense: Optional[MultiDispenseProperties]

    @property
    def aspirate(self) -> AspirateProperties:
        """Aspirate properties."""
        return self._aspirate

    @property
    def dispense(self) -> SingleDispenseProperties:
        """Single dispense properties."""
        return self._dispense

    @property
    def multi_dispense(self) -> Optional[MultiDispenseProperties]:
        """Multi dispense properties."""
        return self._multi_dispense


def _ensure_validated_tip_position(
    tip_position: Union[TipPosition, TipPositionDict],
) -> TipPosition:
    """Given tip position in valid shapes, return an object of TipPosition type."""
    if isinstance(tip_position, TipPosition):
        return tip_position
    elif isinstance(tip_position, dict):
        try:
            pos = SharedDataTipPosition.model_validate(tip_position)
        except ValidationError as e:
            raise ValueError(f"Invalid tip position: {e}") from e
        return _build_tip_position(pos)
    else:
        raise TypeError(
            f"Tip position should be an instance of `TipPosition` or of type `TipPositionDict`, but got {tip_position}"
        )


def _build_tip_position(tip_position: SharedDataTipPosition) -> TipPosition:
    return TipPosition(
        _position_reference=tip_position.positionReference, _offset=tip_position.offset
    )


def _build_delay_properties(
    delay_properties: SharedDataDelayProperties,
) -> DelayProperties:
    if delay_properties.params is not None:
        duration = delay_properties.params.duration
    else:
        duration = None
    return DelayProperties(_enabled=delay_properties.enable, _duration=duration)


def _build_touch_tip_properties(
    touch_tip_properties: SharedDataTouchTipProperties,
) -> TouchTipProperties:
    if touch_tip_properties.params is not None:
        z_offset = touch_tip_properties.params.zOffset
        mm_from_edge = touch_tip_properties.params.mmFromEdge
        speed = touch_tip_properties.params.speed
    else:
        z_offset = None
        mm_from_edge = None
        speed = None
    return TouchTipProperties(
        _enabled=touch_tip_properties.enable,
        _z_offset=z_offset,
        _mm_from_edge=mm_from_edge,
        _speed=speed,
    )


def _build_mix_properties(
    mix_properties: SharedDataMixProperties,
) -> MixProperties:
    if mix_properties.params is not None:
        repetitions = mix_properties.params.repetitions
        volume = mix_properties.params.volume
    else:
        repetitions = None
        volume = None
    return MixProperties(
        _enabled=mix_properties.enable, _repetitions=repetitions, _volume=volume
    )


def _build_blowout_properties(
    blowout_properties: SharedDataBlowoutProperties,
) -> BlowoutProperties:
    if blowout_properties.params is not None:
        location = blowout_properties.params.location
        flow_rate = blowout_properties.params.flowRate
        blowout_position = blowout_properties.params.blowoutPosition
    else:
        location = None
        flow_rate = None
        blowout_position = None
    return BlowoutProperties(
        _enabled=blowout_properties.enable,
        _location=location,
        _flow_rate=flow_rate,
        _blowout_position=_build_tip_position(blowout_position)
        if blowout_position is not None
        else None,
    )


def _build_submerge(
    submerge_properties: SharedDataSubmerge,
) -> Submerge:
    return Submerge(
        _start_position=_build_tip_position(submerge_properties.startPosition),
        _speed=submerge_properties.speed,
        _delay=_build_delay_properties(submerge_properties.delay),
    )


def _build_retract_aspirate(
    retract_aspirate: SharedDataRetractAspirate,
) -> RetractAspirate:
    return RetractAspirate(
        _end_position=_build_tip_position(retract_aspirate.endPosition),
        _speed=retract_aspirate.speed,
        _air_gap_by_volume=LiquidHandlingPropertyByVolume(
            retract_aspirate.airGapByVolume
        ),
        _touch_tip=_build_touch_tip_properties(retract_aspirate.touchTip),
        _delay=_build_delay_properties(retract_aspirate.delay),
    )


def _build_retract_dispense(
    retract_dispense: SharedDataRetractDispense,
) -> RetractDispense:
    return RetractDispense(
        _end_position=_build_tip_position(retract_dispense.endPosition),
        _speed=retract_dispense.speed,
        _air_gap_by_volume=LiquidHandlingPropertyByVolume(
            retract_dispense.airGapByVolume
        ),
        _blowout=_build_blowout_properties(retract_dispense.blowout),
        _touch_tip=_build_touch_tip_properties(retract_dispense.touchTip),
        _delay=_build_delay_properties(retract_dispense.delay),
    )


def build_aspirate_properties(
    aspirate_properties: SharedDataAspirateProperties,
) -> AspirateProperties:
    return AspirateProperties(
        _submerge=_build_submerge(aspirate_properties.submerge),
        _retract=_build_retract_aspirate(aspirate_properties.retract),
        _aspirate_position=_build_tip_position(aspirate_properties.aspiratePosition),
        _flow_rate_by_volume=LiquidHandlingPropertyByVolume(
            aspirate_properties.flowRateByVolume
        ),
        _correction_by_volume=LiquidHandlingPropertyByVolume(
            aspirate_properties.correctionByVolume
        ),
        _pre_wet=aspirate_properties.preWet,
        _mix=_build_mix_properties(aspirate_properties.mix),
        _delay=_build_delay_properties(aspirate_properties.delay),
    )


def build_single_dispense_properties(
    single_dispense_properties: SharedDataSingleDispenseProperties,
) -> SingleDispenseProperties:
    return SingleDispenseProperties(
        _submerge=_build_submerge(single_dispense_properties.submerge),
        _retract=_build_retract_dispense(single_dispense_properties.retract),
        _dispense_position=_build_tip_position(
            single_dispense_properties.dispensePosition
        ),
        _flow_rate_by_volume=LiquidHandlingPropertyByVolume(
            single_dispense_properties.flowRateByVolume
        ),
        _correction_by_volume=LiquidHandlingPropertyByVolume(
            single_dispense_properties.correctionByVolume
        ),
        _mix=_build_mix_properties(single_dispense_properties.mix),
        _push_out_by_volume=LiquidHandlingPropertyByVolume(
            single_dispense_properties.pushOutByVolume
        ),
        _delay=_build_delay_properties(single_dispense_properties.delay),
    )


def build_multi_dispense_properties(
    multi_dispense_properties: Optional[SharedDataMultiDispenseProperties],
) -> Optional[MultiDispenseProperties]:
    if multi_dispense_properties is None:
        return None
    return MultiDispenseProperties(
        _submerge=_build_submerge(multi_dispense_properties.submerge),
        _retract=_build_retract_dispense(multi_dispense_properties.retract),
        _dispense_position=_build_tip_position(
            multi_dispense_properties.dispensePosition
        ),
        _flow_rate_by_volume=LiquidHandlingPropertyByVolume(
            multi_dispense_properties.flowRateByVolume
        ),
        _correction_by_volume=LiquidHandlingPropertyByVolume(
            multi_dispense_properties.correctionByVolume
        ),
        _conditioning_by_volume=LiquidHandlingPropertyByVolume(
            multi_dispense_properties.conditioningByVolume
        ),
        _disposal_by_volume=LiquidHandlingPropertyByVolume(
            multi_dispense_properties.disposalByVolume
        ),
        _delay=_build_delay_properties(multi_dispense_properties.delay),
    )


def build_transfer_properties(
    transfer_properties: Union[SharedDataTransferProperties, SharedByTipTypeSetting],
) -> TransferProperties:
    if isinstance(transfer_properties, SharedByTipTypeSetting):
        _transfer_properties = SharedDataTransferProperties(
            aspirate=transfer_properties.aspirate,
            singleDispense=transfer_properties.singleDispense,
            multiDispense=transfer_properties.multiDispense,
        )
    else:
        _transfer_properties = transfer_properties
    return TransferProperties(
        _aspirate=build_aspirate_properties(_transfer_properties.aspirate),
        _dispense=build_single_dispense_properties(_transfer_properties.singleDispense),
        _multi_dispense=build_multi_dispense_properties(
            _transfer_properties.multiDispense
        ),
    )



from opentrons.config import defaults_ot2
from opentrons.config.types import CurrentDict


def test_load_currents() -> None:
    legacy = {"X": 2.0, "Y": 0.5, "Z": 0.2, "A": 0.1, "B": 0.5, "C": 0.7}
    default: CurrentDict = {
        "default": {"X": 0.1, "Y": 0.3, "Z": 0.1, "A": 0.2, "B": 0.1, "C": 0.2},
        "B": {"X": 0.2, "Y": 0.1, "Z": 0.5, "A": 0.6, "B": 0.7, "C": 0.8},
        "2.1": {"X": 1, "Y": 2, "Z": 3, "A": 4, "B": 5, "C": 7},
    }
    default_different_vals = {
        "default": {"X": 1, "Y": 2, "Z": 3, "A": 4, "B": 5, "C": 6},
        "2.1": {"X": 0, "Y": 1, "Z": 2, "A": 3, "B": 4, "C": 5},
    }
    from_legacy = {"default": default["default"], "B": default["B"], "2.1": legacy}
    assert defaults_ot2._build_hw_versioned_current_dict(legacy, default) == from_legacy
    assert (
        defaults_ot2._build_hw_versioned_current_dict(default_different_vals, default)
        == default_different_vals
    )
    assert defaults_ot2._build_hw_versioned_current_dict(None, default) == default



import pytest

from opentrons.hardware_control.ot3api import OT3API


@pytest.mark.ot2_only
async def test_flex_simulator_always_importable() -> None:
    api = await OT3API.build_hardware_simulator()
    assert isinstance(api, OT3API)



# noqa: D100

from opentrons.protocol_engine.state._axis_aligned_bounding_box import (
    AxisAlignedBoundingBox3D,
)
from opentrons.types import Point


def test_from_corners() -> None:
    """Test the `from_corners()` constructor."""
    result = AxisAlignedBoundingBox3D.from_corners(Point(1, -2, 3), Point(4, -5, 6))
    assert result.min_x == 1
    assert result.max_x == 4
    assert result.min_y == -5
    assert result.max_y == -2
    assert result.min_z == 3
    assert result.max_z == 6


def test_dimensions() -> None:
    """Test the dimension properties."""
    result = AxisAlignedBoundingBox3D.from_corners(Point(1, -2, 3), Point(9, -8, 7))
    assert result.x_dimension == 8
    assert result.y_dimension == 6
    assert result.z_dimension == 4



import pytest
from mock import AsyncMock

from opentrons.drivers.asyncio.communication.serial_connection import (
    AsyncResponseSerialConnection,
)
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.heater_shaker import driver
from opentrons.drivers.types import RPM, HeaterShakerLabwareLatchStatus, Temperature


@pytest.fixture
def connection() -> AsyncMock:
    return AsyncMock(spec=AsyncResponseSerialConnection)


@pytest.fixture
def subject(connection: AsyncMock) -> driver.HeaterShakerDriver:
    connection.send_command.return_value = ""
    return driver.HeaterShakerDriver(connection)


async def test_open_lock(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should send an open plate lock command"""
    connection.send_command.return_value = "M242 ok\n"
    await subject.open_labware_latch()
    expected = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode(
        gcode="M242"
    )
    connection.send_command.assert_called_once_with(command=expected, retries=0)


async def test_close_lock(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should send a close plate lock command"""
    connection.send_command.return_value = "M243 ok\n"
    await subject.close_labware_latch()
    expected = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode(
        gcode="M243"
    )
    connection.send_command.assert_called_once_with(command=expected, retries=0)


async def test_get_lock_status(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should send a get plate lock status command and parse response."""
    connection.send_command.return_value = "M241 STATUS:IDLE_UNKNOWN ok\n"

    response = await subject.get_labware_latch_status()

    expected = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode(
        gcode="M241"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=0)
    assert response == HeaterShakerLabwareLatchStatus.IDLE_UNKNOWN


async def test_set_temperature(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should send a set temperature command"""
    connection.send_command.return_value = "M104 ok\n"
    await subject.set_temperature(temperature=65.02)
    expected = (
        CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR)
        .add_gcode(gcode="M104")
        .add_float(prefix="S", value=65.02, precision=2)
    )
    connection.send_command.assert_called_once_with(command=expected, retries=0)


async def test_get_temperature(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should send a get temperature command and parse the response"""

    connection.send_command.return_value = "M105 T:91.25 C:54.02 ok\n"
    response = await subject.get_temperature()
    expected = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode("M105")
    connection.send_command.assert_called_once_with(command=expected, retries=0)
    assert response == Temperature(current=54.02, target=91.25)


async def test_set_rpm(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should send a set rpm command"""
    connection.send_command.return_value = "M3 ok\n"
    await subject.set_rpm(2500)
    expected = (
        CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR)
        .add_gcode(gcode="M3")
        .add_int(prefix="S", value=2500)
    )
    connection.send_command.assert_called_once_with(command=expected, retries=0)


async def test_get_rpm(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should send a get rpm command and parse the response"""
    connection.send_command.return_value = "M123 T:2200 C:2100 ok\n"
    response = await subject.get_rpm()
    expected = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode("M123")
    connection.send_command.assert_called_once_with(command=expected, retries=0)
    assert response == RPM(current=2100, target=2200)


async def test_home(subject: driver.HeaterShakerDriver, connection: AsyncMock) -> None:
    """It should send a home command"""
    connection.send_command.return_value = "G28 ok\n"
    await subject.home()
    expected = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode("G28")
    connection.send_command.assert_called_once_with(command=expected, retries=0)


async def test_deactivate_heater(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should send a deactivate-heater command"""
    connection.send_command.return_value = "M106 ok\n"
    await subject.deactivate_heater()
    expected = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode("M106")
    connection.send_command.assert_called_once_with(command=expected, retries=0)


async def test_get_device_info(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should send a get device info command"""
    connection.send_command.return_value = (
        "M115 FW:21.2.1 HW:A SerialNo:TC2101010A2 ok\n"
    )
    response = await subject.get_device_info()
    assert response == {"serial": "TC2101010A2", "model": "A", "version": "21.2.1"}
    device_info = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode(
        gcode="M115"
    )
    reset_reason = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode(
        gcode="M114"
    )
    connection.send_command.assert_any_call(command=device_info, retries=0)
    connection.send_command.assert_called_with(command=reset_reason, retries=0)


async def test_enter_bootloader(
    subject: driver.HeaterShakerDriver, connection: AsyncMock
) -> None:
    """It should enter the bootloader"""
    connection.send_dfu_command.return_value = None
    await subject.enter_programming_mode()
    expected = CommandBuilder(terminator=driver.HS_COMMAND_TERMINATOR).add_gcode(
        gcode="dfu"
    )
    connection.send_dfu_command.assert_called_once_with(command=expected)
    connection.close.assert_called_once()



from typing import AsyncGenerator

import pytest
from mock import MagicMock

from opentrons.config.defaults_ot2 import (
    DEFAULT_GANTRY_STEPS_PER_MM,
    DEFAULT_PIPETTE_CONFIGS,
)
from opentrons.config.robot_configs import build_config_ot2
from opentrons.drivers.smoothie_drivers import SmoothieDriver
from opentrons.drivers.types import MoveSplit
from opentrons.hardware_control.emulation.settings import Settings


@pytest.fixture
async def subject(emulator_settings: Settings) -> AsyncGenerator[SmoothieDriver, None]:
    """Smoothie driver connected to emulator."""
    d = await SmoothieDriver.build(
        port=f"socket://127.0.0.1:{emulator_settings.smoothie.port}",
        config=build_config_ot2({}),
    )
    try:
        yield d
    finally:
        await d.disconnect()


@pytest.fixture
def spy(subject: SmoothieDriver) -> MagicMock:
    """Attach a spy to gcode sender."""
    assert subject._connection is not None
    spy = MagicMock(wraps=subject._connection.send_data)
    subject._connection.send_data = spy  # type: ignore[method-assign]
    return spy


async def test_dwell_and_activate_axes(subject: SmoothieDriver, spy: MagicMock) -> None:
    subject.activate_axes("X")
    await subject._set_saved_current()
    subject.dwell_axes("X")
    await subject._set_saved_current()
    subject.activate_axes("XYBC")
    await subject._set_saved_current()
    subject.dwell_axes("XC")
    await subject._set_saved_current()
    subject.dwell_axes("BCY")
    await subject._set_saved_current()
    expected = [
        "M907 A0.1 B0.05 C0.05 X1.25 Y0.3 Z0.1 G4 P0.005",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
        "M907 A0.1 B0.05 C0.05 X1.25 Y1.25 Z0.1 G4 P0.005",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y1.25 Z0.1 G4 P0.005",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]

    assert command_log == expected


async def test_disable_motor(subject: SmoothieDriver, spy: MagicMock) -> None:
    await subject.disengage_axis("X")
    await subject.disengage_axis("XYZ")
    await subject.disengage_axis("ABCD")
    expected = [
        "M18 X",
        "M400",
        "M18 XYZ",
        "M400",
        "M18 ABC",
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == expected


async def test_plunger_commands(subject: SmoothieDriver, spy: MagicMock) -> None:
    await subject.home()
    expected = [
        "M907 A0.8 B0.05 C0.05 X0.3 Y0.3 Z0.8 G4 P0.005 G28.2 ABCZ",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
        "M203.1 Y50",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.8 Z0.1 G4 P0.005 G91 G0 Y-28 G0 Y10 G90",
        "M400",
        "M203.1 X80",
        "M400",
        "M907 A0.1 B0.05 C0.05 X1.25 Y0.3 Z0.1 G4 P0.005 G28.2 X",
        "M400",
        "M203.1 A125 B40 C40 X600 Y400 Z125",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
        "M203.1 Y80",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y1.25 Z0.1 G4 P0.005 G28.2 Y",
        "M400",
        "M203.1 Y8",
        "M400",
        "G91 G0 Y-3 G90",
        "M400",
        "G28.2 Y",
        "M400",
        "G91 G0 Y-3 G90",
        "M400",
        "M203.1 A125 B40 C40 X600 Y400 Z125",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
        "M114.2",
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == expected

    spy.reset_mock()

    await subject.move({"X": 0, "Y": 1.123456, "Z": 2, "A": 3})
    expected = [
        "M907 A0.8 B0.05 C0.05 X1.25 Y1.25 Z0.8 G4 P0.005 G0 A3 X0 Y1.123 Z2",
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == expected

    spy.reset_mock()

    await subject.move({"B": 2})
    expected = [
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005 G0 B2",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == expected

    spy.reset_mock()

    await subject.move(
        {"X": 10.987654321, "Y": 2.12345678, "Z": 2.5, "A": 3.5, "B": 4.25, "C": 5.55}
    )
    expected = [
        # Set active axes high
        "M907 A0.8 B0.05 C0.05 X1.25 Y1.25 Z0.8 G4 P0.005"
        " G0 A3.5 B4.55 C5.55 X10.988 Y2.123 Z2.5 G0 B4.25",
        "M400",
        # Set plunger current low
        "M907 A0.8 B0.05 C0.05 X1.25 Y1.25 Z0.8 G4 P0.005",
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == expected


async def test_move_with_split(subject: SmoothieDriver, spy: MagicMock) -> None:
    await subject.home()
    spy.reset_mock()

    subject.configure_splits_for(
        {
            "B": MoveSplit(
                split_distance=1,
                split_current=1.75,
                split_speed=1,
                after_time=1800,
                fullstep=True,
            ),
            "C": MoveSplit(
                split_distance=1,
                split_current=1.75,
                split_speed=1,
                after_time=1800,
                fullstep=True,
            ),
        }
    )
    subject._steps_per_mm = {"B": 1.0, "C": 1.0}

    await subject.move({"X": 0, "Y": 1.123456, "Z": 2, "C": 3})
    expected = [
        "M55 M92 C0.03125 G4 P0.01 G0 F60 M907 A0.1 B0.05 C1.75 X1.25 Y1.25"
        " Z0.8 G4 P0.005",
        "M400",
        "G0 C18.0",
        "M400",
        "M54 M92 C1.0 G4 P0.01",
        "M400",
        "G0 F24000 M907 A0.1 B0.05 C0.05 X1.25 Y1.25 Z0.8 G4 P0.005 G0 C3 X0 Y1.123 Z2",
        "M400",
        "M907 A0.1 B0.05 C0.05 X1.25 Y1.25 Z0.8 G4 P0.005",
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == expected

    spy.reset_mock()

    await subject.move({"B": 2})
    expected = [
        "M53 M92 B0.03125 G4 P0.01 G0 F60 M907 A0.1 B1.75 C0.05 X0.3 Y0.3 Z0.1"
        " G4 P0.005",
        "M400",
        "G0 B18.0",
        "M400",
        "M52 M92 B1.0 G4 P0.01",
        "M400",
        "G0 F24000 M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005 G0 B2",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == expected


async def test_set_active_current(subject: SmoothieDriver, spy: MagicMock) -> None:
    await subject.home()
    spy.reset_mock()

    subject.set_active_current({"X": 2, "Y": 2, "Z": 2, "A": 2, "B": 2, "C": 2})
    subject.set_dwelling_current({"X": 0, "Y": 0, "Z": 0, "A": 0, "B": 0, "C": 0})

    await subject.move({"X": 0, "Y": 0, "Z": 0, "A": 0, "B": 0, "C": 0})
    await subject.move({"B": 1, "C": 1})
    subject.set_active_current({"B": 0.42, "C": 0.42})
    await subject.home("BC")
    expected = [
        # move all
        "M907 A2 B2 C2 X2 Y2 Z2 G4 P0.005 G0 A0 B0 C0 X0 Y0 Z0",
        "M400",
        "M907 A2 B0 C0 X2 Y2 Z2 G4 P0.005",  # disable BC axes
        "M400",
        # move BC
        "M907 A0 B2 C2 X0 Y0 Z0 G4 P0.005 G0 B1.3 C1.3 G0 B1 C1",
        "M400",
        "M907 A0 B0 C0 X0 Y0 Z0 G4 P0.005",  # disable BC axes
        "M400",
        "M907 A0 B0.42 C0.42 X0 Y0 Z0 G4 P0.005 G28.2 BC",  # home BC
        "M400",
        "M907 A0 B0 C0 X0 Y0 Z0 G4 P0.005",  # dwell all axes after home
        "M400",
        "M114.2",  # update the position
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == expected


async def test_steps_per_mm(subject: SmoothieDriver, spy: MagicMock) -> None:
    expected = {
        **DEFAULT_GANTRY_STEPS_PER_MM,
        "B": DEFAULT_PIPETTE_CONFIGS["stepsPerMM"],
        "C": DEFAULT_PIPETTE_CONFIGS["stepsPerMM"],
    }
    assert subject.steps_per_mm == expected
    await subject.update_steps_per_mm({"Z": 450})
    expected["Z"] = 450
    assert subject.steps_per_mm == expected

    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == ["M92 Z450", "M400"]


async def test_pipette_configs(subject: SmoothieDriver, spy: MagicMock) -> None:
    res = await subject.update_pipette_config(
        "Z", {"home": 175, "debounce": 12, "max_travel": 13, "retract": 14}
    )
    expected_return = {
        "Z": {"home": 175, "debounce": 12, "max_travel": 13, "retract": 14}
    }
    assert res == expected_return

    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == [
        "M365.0 Z175",
        "M400",
        "M365.2 O12",
        "M400",
        "M365.1 Z13",
        "M400",
        "M365.3 Z14",
        "M400",
    ]


async def test_set_acceleration(subject: SmoothieDriver, spy: MagicMock) -> None:
    await subject.set_acceleration({"X": 1, "Y": 2, "Z": 3, "A": 4, "B": 5, "C": 6})
    subject.push_acceleration()
    await subject.pop_acceleration()
    await subject.set_acceleration(
        {"X": 10, "Y": 20, "Z": 30, "A": 40, "B": 50, "C": 60}
    )
    await subject.pop_acceleration()

    expected = [
        "M204 S10000 A4 B5 C6 X1 Y2 Z3",
        "M400",
        "M204 S10000 A4 B5 C6 X1 Y2 Z3",
        "M400",
        "M204 S10000 A40 B50 C60 X10 Y20 Z30",
        "M400",
        "M204 S10000 A4 B5 C6 X1 Y2 Z3",
        "M400",
    ]
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == expected


async def test_read_and_write_pipettes(subject: SmoothieDriver) -> None:
    test_id = "TestsRock!!"
    test_model = "TestPipette"
    await subject.write_pipette_id("left", test_id)
    read_id = await subject.read_pipette_id("left")
    assert read_id == test_id

    await subject.write_pipette_model("left", test_model)
    read_model = await subject.read_pipette_model("left")
    assert read_model == test_model + "_v1"


async def test_fast_home(subject: SmoothieDriver, spy: MagicMock) -> None:
    await subject.fast_home(axis="X", safety_margin=12)

    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == [
        "M907 A0.1 B0.05 C0.05 X1.25 Y0.3 Z0.1 G4 P0.005 G0 X406.0",
        "M400",
        "M203.1 Y50",
        "M400",
        "M907 A0.1 B0.05 C0.05 X1.25 Y0.8 Z0.1 G4 P0.005 G91 G0 Y-28 G0 Y10 G90",
        "M400",
        "M203.1 X80",
        "M400",
        "M907 A0.1 B0.05 C0.05 X1.25 Y0.3 Z0.1 G4 P0.005 G28.2 X",
        "M400",
        "M203.1 A125 B40 C40 X600 Y400 Z125",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
        "M114.2",
        "M400",
    ]


async def test_fast_home_two_axes_plus_two_invalid_axes(
    subject: SmoothieDriver, spy: MagicMock
) -> None:
    await subject.fast_home(axis="XYJK", safety_margin=12)

    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == [
        "M907 A0.1 B0.05 C0.05 X1.25 Y1.25 Z0.1 G4 P0.005 G0 X406.0 Y341.0",
        "M400",
        "M203.1 Y50",
        "M400",
        "M907 A0.1 B0.05 C0.05 X1.25 Y0.8 Z0.1 G4 P0.005 G91 G0 Y-28 G0 Y10 G90",
        "M400",
        "M203.1 X80",
        "M400",
        "M907 A0.1 B0.05 C0.05 X1.25 Y0.3 Z0.1 G4 P0.005 G28.2 X",
        "M400",
        "M203.1 A125 B40 C40 X600 Y400 Z125",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
        "M203.1 Y80",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y1.25 Z0.1 G4 P0.005 G28.2 Y",
        "M400",
        "M203.1 Y8",
        "M400",
        "G91 G0 Y-3 G90",
        "M400",
        "G28.2 Y",
        "M400",
        "G91 G0 Y-3 G90",
        "M400",
        "M203.1 A125 B40 C40 X600 Y400 Z125",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
        "M114.2",
        "M400",
    ]


async def test_update_homing_flags(subject: SmoothieDriver, spy: MagicMock) -> None:
    await subject.update_homed_flags()
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == ["G28.6", "M400"]


async def test_update_pipette_config(subject: SmoothieDriver, spy: MagicMock) -> None:
    await subject.update_pipette_config(
        "X", {"retract": 2, "debounce": 3, "max_travel": 4, "home": 5}
    )
    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == [
        "M365.3 X2",
        "M400",
        "M365.2 O3",
        "M400",
        "M365.1 X4",
        "M400",
        "M365.0 X5",
        "M400",
    ]


async def test_do_relative_splits_during_home_for(
    subject: SmoothieDriver, spy: MagicMock
) -> None:
    """Test command structure when a split configuration is present."""
    subject.configure_splits_for(
        {
            "B": MoveSplit(
                split_distance=1,
                split_current=1.75,
                split_speed=1,
                after_time=1800,
                fullstep=True,
            )
        }
    )
    subject._steps_per_mm = {"B": 1.0, "C": 1.0}

    await subject._do_relative_splits_during_home_for("BC")

    command_log = [x.kwargs["data"].strip() for x in spy.call_args_list]
    assert command_log == [
        "M53 M55 M92 B0.03125 C0.03125 G4 P0.01 M907 B1.75 G4 P0.005 G0 F60 G91",
        "M400",
        "G0 B-1",
        "M400",
        "G90 M52 M54 M92 B1.0 C1.0 G4 P0.01 G0 F24000",
        "M400",
    ]



"""Tests for hash_command_params."""

import pytest

from opentrons.protocol_engine import CommandIntent, commands
from opentrons.protocol_engine.commands.hash_command_params import (
    hash_protocol_command_params,
)


def test_equivalent_commands() -> None:
    """Equivalent commands should have the same hash."""
    a = commands.BlowOutInPlaceCreate(
        params=commands.BlowOutInPlaceParams(
            pipetteId="abc123",
            flowRate=123,
        )
    )
    b = commands.WaitForDurationCreate(
        params=commands.WaitForDurationParams(seconds=123)
    )
    c = commands.WaitForDurationCreate(
        params=commands.WaitForDurationParams(seconds=123)
    )

    assert hash_protocol_command_params(b, None) == hash_protocol_command_params(
        c, None
    )

    a_hash = hash_protocol_command_params(a, None)
    assert hash_protocol_command_params(b, a_hash) == hash_protocol_command_params(
        c, a_hash
    )


def test_nonequivalent_commands() -> None:
    """Nonequivalent commands should have different hashes."""
    a = commands.BlowOutInPlaceCreate(
        params=commands.BlowOutInPlaceParams(
            pipetteId="abc123",
            flowRate=123,
        ),
        intent=CommandIntent.PROTOCOL,
    )
    b = commands.WaitForDurationCreate(
        params=commands.WaitForDurationParams(seconds=123)
    )

    assert hash_protocol_command_params(a, None) != hash_protocol_command_params(
        b, None
    )


def test_repeated_commands() -> None:
    """Repeated commands should hash differently, even though they're equivalent in isolation."""
    a = commands.WaitForDurationCreate(
        params=commands.WaitForDurationParams(seconds=123),
        intent=CommandIntent.PROTOCOL,
    )
    b = commands.WaitForDurationCreate(
        params=commands.WaitForDurationParams(seconds=123),
        intent=CommandIntent.PROTOCOL,
    )

    a_hash = hash_protocol_command_params(a, None)
    b_hash = hash_protocol_command_params(b, a_hash)
    assert a_hash != b_hash


@pytest.mark.parametrize("command_intent", [CommandIntent.SETUP, CommandIntent.FIXIT])
def test_setup_and_fixit_command(command_intent: CommandIntent) -> None:
    """Setup and fixit commands should always skip hashing."""
    setup_command = commands.WaitForDurationCreate(
        params=commands.WaitForDurationParams(seconds=123),
        intent=command_intent,
    )
    assert hash_protocol_command_params(setup_command, None) is None



"""Test Flex Stacker fill command implementation."""

from typing import cast

import pytest
from decoy import Decoy

from opentrons_shared_data.errors.exceptions import CommandPreconditionViolated
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.commands.flex_stacker.fill import (
    FillImpl,
    FillParams,
    FillResult,
)
from opentrons.protocol_engine.errors import (
    FlexStackerLabwarePoolNotYetDefinedError,
    ModuleNotLoadedError,
)
from opentrons.protocol_engine.execution import EquipmentHandler, RunControlHandler
from opentrons.protocol_engine.execution.equipment import LoadedLabwarePoolData
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    BatchLabwareLocationUpdate,
    BatchLoadedLabwareUpdate,
    FlexStackerStateUpdate,
    LabwareLidUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    OFF_DECK_LOCATION,
    SYSTEM_LOCATION,
    DeckSlotLocation,
    InStackerHopperLocation,
    LoadedLabware,
    NotOnDeckLocationSequenceComponent,
    StackerFillEmptyStrategy,
    StackerStoredLabwareGroup,
)
from opentrons.types import DeckSlotName


@pytest.fixture
def subject(
    state_view: StateView,
    run_control: RunControlHandler,
    model_utils: ModelUtils,
    equipment: EquipmentHandler,
) -> FillImpl:
    """A FillImpl for testing."""
    return FillImpl(
        state_view=state_view,
        run_control=run_control,
        equipment=equipment,
        model_utils=model_utils,
    )


def _contained_labware(count: int) -> list[StackerStoredLabwareGroup]:
    return [
        StackerStoredLabwareGroup(
            primaryLabwareId=f"primary-id-{i + 1}",
            adapterLabwareId=None,
            lidLabwareId=None,
        )
        for i in range(count)
    ]


@pytest.mark.parametrize(
    "current_stored,count_param,max_pool_count",
    [
        pytest.param(_contained_labware(3), 3, 3, id="already-full"),
        pytest.param(_contained_labware(0), 4, 3, id="empty-fill-more-than-max"),
        pytest.param(_contained_labware(1), 3, 3, id="nonempty-fill-more-than-max"),
    ],
)
async def test_fill_by_count_exceeding_max(
    decoy: Decoy,
    state_view: StateView,
    model_utils: ModelUtils,
    equipment: EquipmentHandler,
    subject: FillImpl,
    current_stored: list[StackerStoredLabwareGroup],
    count_param: int,
    max_pool_count: int,
    flex_50uL_tiprack: LabwareDefinition,
) -> None:
    """It should fill a valid stacker's labware pool."""
    module_id = "some-module-id"
    stacker_state = FlexStackerSubState(
        module_id=cast(FlexStackerId, module_id),
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=current_stored,
        max_pool_count=max_pool_count,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        stacker_state
    )
    params = FillParams(
        moduleId=module_id,
        count=count_param,
        message="some-message",
        strategy=StackerFillEmptyStrategy.LOGICAL,
    )
    with pytest.raises(CommandPreconditionViolated):
        await subject.execute(params)


@pytest.mark.parametrize(
    "current_stored,count_param,max_pool_count,expected_new_labware",
    [
        pytest.param([], 3, 3, 3, id="empty-to-full"),
        pytest.param(
            _contained_labware(1),
            None,
            3,
            2,
            id="default-count",
        ),
    ],
)
async def test_fill_by_count_happypath(
    decoy: Decoy,
    state_view: StateView,
    model_utils: ModelUtils,
    equipment: EquipmentHandler,
    subject: FillImpl,
    current_stored: list[StackerStoredLabwareGroup],
    count_param: int | None,
    max_pool_count: int,
    expected_new_labware: int,
    flex_50uL_tiprack: LabwareDefinition,
) -> None:
    """It should fill a valid stacker's labware pool."""
    module_id = "some-module-id"
    stacker_state = FlexStackerSubState(
        module_id=cast(FlexStackerId, module_id),
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=current_stored,
        max_pool_count=max_pool_count,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        stacker_state
    )
    params = FillParams(
        moduleId=module_id,
        count=count_param,
        message="some-message",
        strategy=StackerFillEmptyStrategy.LOGICAL,
    )
    primary_ids = iter([f"new-primary-{i + 1}" for i in range(expected_new_labware)])
    decoy.when(model_utils.generate_id()).then_do(lambda: next(primary_ids))
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(flex_50uL_tiprack)
    ).then_return("opentrons/opentrons_flex_96_filtertiprack_50ul/1")
    for i in range(expected_new_labware):
        decoy.when(state_view.labware.known(f"new-primary-{i + 1}")).then_return(False)
        decoy.when(
            await equipment.load_labware_pool_from_definitions(
                pool_primary_definition=flex_50uL_tiprack,
                pool_adapter_definition=None,
                pool_lid_definition=None,
                location=InStackerHopperLocation(moduleId="some-module-id"),
                primary_id=f"new-primary-{i + 1}",
                adapter_id=None,
                lid_id=None,
            )
        ).then_return(
            LoadedLabwarePoolData(
                primary_labware=LoadedLabware(
                    id=f"new-primary-{i + 1}",
                    loadName="loadname",
                    definitionUri="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
                    location=InStackerHopperLocation(moduleId="some-module-id"),
                    lid_id=None,
                    offsetId=None,
                    displayName=None,
                ),
                adapter_labware=None,
                lid_labware=None,
            )
        )
    result = await subject.execute(params)
    added_labware = [
        StackerStoredLabwareGroup(
            primaryLabwareId=f"new-primary-{i + 1}",
            adapterLabwareId=None,
            lidLabwareId=None,
        )
        for i in range(expected_new_labware)
    ]
    new_stored_labware = current_stored + added_labware
    assert result.state_update == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id=module_id, contained_labware_bottom_first=new_stored_labware
        ),
        batch_loaded_labware=BatchLoadedLabwareUpdate(
            new_locations_by_id={
                f"new-primary-{i + 1}": InStackerHopperLocation(
                    moduleId="some-module-id"
                )
                for i in range(expected_new_labware)
            },
            offset_ids_by_id={
                f"new-primary-{i + 1}": None for i in range(expected_new_labware)
            },
            display_names_by_id={
                f"new-primary-{i + 1}": None for i in range(expected_new_labware)
            },
            definitions_by_id={
                f"new-primary-{i + 1}": flex_50uL_tiprack
                for i in range(expected_new_labware)
            },
        ),
        labware_lid=LabwareLidUpdate(parent_labware_ids=[], lid_ids=[]),
    )
    assert result.public.storedLabware == new_stored_labware
    assert result.public.addedLabware == added_labware
    assert result.public == FillResult(
        count=(len(current_stored) + expected_new_labware),
        storedLabware=new_stored_labware,
        addedLabware=added_labware,
        primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
        adapterLabwareURI=None,
        lidLabwareURI=None,
        originalPrimaryLabwareLocationSequences=[
            [NotOnDeckLocationSequenceComponent(logicalLocationName=SYSTEM_LOCATION)]
            for _ in added_labware
        ],
        originalAdapterLabwareLocationSequences=None,
        originalLidLabwareLocationSequences=None,
        newPrimaryLabwareLocationSequences=[
            [InStackerHopperLocation(moduleId="some-module-id")] for _ in added_labware
        ],
        newAdapterLabwareLocationSequences=None,
        newLidLabwareLocationSequences=None,
    )


async def test_fill_requires_stacker(
    decoy: Decoy, state_view: StateView, subject: FillImpl
) -> None:
    """It should require a stacker."""
    decoy.when(state_view.modules.get_flex_stacker_substate("asda")).then_raise(
        ModuleNotLoadedError(module_id="asda")
    )
    with pytest.raises(ModuleNotLoadedError):
        await subject.execute(
            FillParams(
                moduleId="asda",
                strategy=StackerFillEmptyStrategy.LOGICAL,
                message="blah",
                count=3,
            )
        )


async def test_fill_requires_constrained_pool(
    decoy: Decoy, state_view: StateView, subject: FillImpl
) -> None:
    """It should require a constrained labware pool."""
    module_id = "module-id"
    stacker_state = FlexStackerSubState(
        module_id=cast(FlexStackerId, module_id),
        pool_primary_definition=None,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(3),
        max_pool_count=0,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        stacker_state
    )
    decoy.when(state_view.modules.get_location(module_id)).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_A3)
    )
    with pytest.raises(
        FlexStackerLabwarePoolNotYetDefinedError,
        match=".*The Flex Stacker in.*A3.*has not been configured yet and cannot be filled.",
    ):
        await subject.execute(
            FillParams(
                moduleId=module_id,
                count=2,
                message="hello",
                strategy=StackerFillEmptyStrategy.LOGICAL,
            )
        )


async def test_pause_strategy_pauses(
    decoy: Decoy,
    state_view: StateView,
    run_control: RunControlHandler,
    model_utils: ModelUtils,
    subject: FillImpl,
    flex_50uL_tiprack: LabwareDefinition,
) -> None:
    """It should pause the system when the pause strategy is used."""
    current_count = 1
    max_pool_count = 6
    module_id = "some-module-id"
    beginning_contained_labware = _contained_labware(current_count)
    stacker_state = FlexStackerSubState(
        module_id=cast(FlexStackerId, module_id),
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=beginning_contained_labware,
        max_pool_count=max_pool_count,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        stacker_state
    )
    new_labware = [
        StackerStoredLabwareGroup(
            primaryLabwareId="new-primary", adapterLabwareId=None, lidLabwareId=None
        )
    ]
    params = FillParams(
        moduleId=module_id,
        labwareToStore=new_labware,
        message="some-message",
        strategy=StackerFillEmptyStrategy.MANUAL_WITH_PAUSE,
    )
    decoy.when(state_view.labware.known("new-primary")).then_return(True)
    decoy.when(state_view.geometry.get_location_sequence("new-primary")).then_return(
        [NotOnDeckLocationSequenceComponent(logicalLocationName=OFF_DECK_LOCATION)]
    )
    decoy.when(state_view.labware.get_location("new-primary")).then_return(
        OFF_DECK_LOCATION
    )
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(flex_50uL_tiprack)
    ).then_return("opentrons/opentrons_flex_96_filtertiprack_50ul/1")
    new_contained_labware = beginning_contained_labware + new_labware
    result = await subject.execute(params)
    assert result.state_update == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id=module_id, contained_labware_bottom_first=new_contained_labware
        ),
        batch_labware_location=BatchLabwareLocationUpdate(
            new_locations_by_id={
                "new-primary": InStackerHopperLocation(moduleId="some-module-id")
            },
            new_offset_ids_by_id={"new-primary": None},
        ),
    )
    assert result.public == FillResult(
        count=2,
        primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
        storedLabware=new_contained_labware,
        addedLabware=new_labware,
        originalPrimaryLabwareLocationSequences=[
            [NotOnDeckLocationSequenceComponent(logicalLocationName=OFF_DECK_LOCATION)]
        ],
        newPrimaryLabwareLocationSequences=[
            [InStackerHopperLocation(moduleId="some-module-id")]
        ],
    )
    decoy.verify(await run_control.wait_for_resume())



from .abstract import AbstractThermocyclerDriver
from .driver import ThermocyclerDriver, ThermocyclerDriverFactory, ThermocyclerDriverV2
from .simulator import SimulatingDriver

__all__ = [
    "ThermocyclerDriver",
    "ThermocyclerDriverV2",
    "ThermocyclerDriverFactory",
    "SimulatingDriver",
    "AbstractThermocyclerDriver",
]



from typing import List, Optional, Tuple, cast

import pytest
from decoy import Decoy

from opentrons_hardware.hardware_control.estop.detector import (
    EstopDetector,
    EstopSummary,
)

from opentrons.hardware_control.backends.estop_state import EstopStateMachine
from opentrons.hardware_control.types import (
    EstopAttachLocation,
    EstopPhysicalStatus,
    EstopState,
    EstopStateNotification,
    HardwareEvent,
)


@pytest.fixture
def initial_state() -> "EstopSummary":
    return EstopSummary(left_detected=True, right_detected=True, engaged=False)


@pytest.fixture
def mock_estop_detector(decoy: Decoy, initial_state: "EstopSummary") -> "EstopDetector":
    """Create a mocked estop detector."""
    mock = decoy.mock(cls=EstopDetector)
    decoy.when(mock.status).then_return(initial_state)
    return mock


@pytest.fixture
def subject(mock_estop_detector: "EstopDetector") -> "EstopStateMachine":
    return EstopStateMachine(detector=mock_estop_detector)


@pytest.mark.ot3_only
async def test_estop_state_no_detector(
    mock_estop_detector: "EstopDetector", decoy: Decoy
) -> None:
    """Test that the estop state machine works without a detector."""
    subject = EstopStateMachine(detector=None)
    assert subject.state == EstopState.DISENGAGED
    assert (
        subject.get_physical_status(EstopAttachLocation.LEFT)
        == EstopPhysicalStatus.DISENGAGED
    )
    assert (
        subject.get_physical_status(EstopAttachLocation.RIGHT)
        == EstopPhysicalStatus.DISENGAGED
    )

    decoy.when(mock_estop_detector.status).then_return(
        EstopSummary(left_detected=False, right_detected=True, engaged=True)
    )

    subject.subscribe_to_detector(detector=mock_estop_detector)

    assert cast(EstopState, subject.state) == EstopState.PHYSICALLY_ENGAGED
    assert (
        subject.get_physical_status(EstopAttachLocation.LEFT)
        == EstopPhysicalStatus.NOT_PRESENT
    )
    assert (
        subject.get_physical_status(EstopAttachLocation.RIGHT)
        == EstopPhysicalStatus.ENGAGED
    )

    # Check that adding a second listener will wipe out the first one
    subject.subscribe_to_detector(detector=mock_estop_detector)

    decoy.verify(
        mock_estop_detector.add_listener(subject.detector_listener),
        mock_estop_detector.remove_listener(subject.detector_listener),
        mock_estop_detector.add_listener(subject.detector_listener),
    )


@pytest.mark.ot3_only
async def test_estop_state_listener(
    subject: "EstopStateMachine", mock_estop_detector: "EstopDetector", decoy: Decoy
) -> None:
    """Test that the state machine broadcasts to listeners."""
    decoy.verify(mock_estop_detector.add_listener(subject.detector_listener))

    events: List[EstopStateNotification] = []

    def mock_listener(event: HardwareEvent) -> None:
        assert isinstance(event, EstopStateNotification)
        events.append(event)

    subject.add_listener(mock_listener)

    steps: List[Tuple[EstopSummary, EstopState]] = [
        (
            EstopSummary(left_detected=False, right_detected=False, engaged=False),
            EstopState.NOT_PRESENT,
        ),
        (
            EstopSummary(left_detected=True, right_detected=False, engaged=False),
            EstopState.DISENGAGED,
        ),
        (
            EstopSummary(left_detected=True, right_detected=False, engaged=True),
            EstopState.PHYSICALLY_ENGAGED,
        ),
    ]

    prev: Optional[EstopState] = None
    for input, result in steps:
        subject.detector_listener(summary=input)
        assert len(events) == 1
        event = events.pop(0)
        assert event.new_state == result
        assert event.new_state == subject.state
        if prev is not None:
            assert event.old_state == prev
        prev = event.new_state

    # Don't transmit event if nothing changed here
    assert subject.acknowledge_and_clear() == EstopState.PHYSICALLY_ENGAGED
    assert len(events) == 0

    # Now unsubscribe, change state, and assert we didn't get a callback
    subject.remove_listener(mock_listener)

    subject.detector_listener(
        summary=EstopSummary(left_detected=True, right_detected=True, engaged=False)
    )
    assert subject.state == EstopState.LOGICALLY_ENGAGED
    assert len(events) == 0

    subject.add_listener(mock_listener)
    assert subject.acknowledge_and_clear() == EstopState.DISENGAGED
    assert len(events) == 1
    assert events.pop().new_state == EstopState.DISENGAGED


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    argnames=["left_connected", "right_connected"],
    argvalues=[
        [False, False],
        [True, False],
        [False, True],
        [True, True],
    ],
)
async def test_estop_physical_status(
    subject: "EstopStateMachine", left_connected: bool, right_connected: bool
) -> None:
    """Test that physical status maps to expected"""
    summary = EstopSummary(
        left_detected=left_connected, right_detected=right_connected, engaged=False
    )
    subject.detector_listener(summary=summary)
    assert (
        subject.get_physical_status(location=EstopAttachLocation.LEFT)
        == EstopPhysicalStatus.DISENGAGED
        if left_connected
        else EstopPhysicalStatus.NOT_PRESENT
    )
    assert (
        subject.get_physical_status(location=EstopAttachLocation.RIGHT)
        == EstopPhysicalStatus.DISENGAGED
        if right_connected
        else EstopPhysicalStatus.NOT_PRESENT
    )

    summary.engaged = True
    subject.detector_listener(summary=summary)
    assert (
        subject.get_physical_status(location=EstopAttachLocation.LEFT)
        == EstopPhysicalStatus.ENGAGED
        if left_connected
        else EstopPhysicalStatus.NOT_PRESENT
    )
    assert (
        subject.get_physical_status(location=EstopAttachLocation.RIGHT)
        == EstopPhysicalStatus.ENGAGED
        if right_connected
        else EstopPhysicalStatus.NOT_PRESENT
    )


@pytest.mark.ot3_only
@pytest.mark.parametrize("clear_to_disengaged", [True, False])
async def test_estop_state_machine(
    subject: "EstopStateMachine", clear_to_disengaged: bool
) -> None:
    """Test traversal through the state machine."""

    steps: List[Tuple[EstopSummary, EstopState]] = [
        (
            EstopSummary(left_detected=False, right_detected=False, engaged=False),
            EstopState.NOT_PRESENT,
        ),
        (
            EstopSummary(left_detected=True, right_detected=False, engaged=False),
            EstopState.DISENGAGED,
        ),
        (
            EstopSummary(left_detected=True, right_detected=True, engaged=False),
            EstopState.DISENGAGED,
        ),
        (
            EstopSummary(left_detected=False, right_detected=False, engaged=False),
            EstopState.NOT_PRESENT,
        ),
        (
            EstopSummary(left_detected=True, right_detected=True, engaged=False),
            EstopState.DISENGAGED,
        ),
        (
            EstopSummary(left_detected=True, right_detected=True, engaged=True),
            EstopState.PHYSICALLY_ENGAGED,
        ),
        (
            EstopSummary(left_detected=True, right_detected=True, engaged=False),
            EstopState.LOGICALLY_ENGAGED,
        ),
        (
            EstopSummary(left_detected=True, right_detected=False, engaged=True),
            EstopState.PHYSICALLY_ENGAGED,
        ),
        (
            EstopSummary(left_detected=True, right_detected=False, engaged=False),
            EstopState.LOGICALLY_ENGAGED,
        ),
        (
            EstopSummary(
                left_detected=False, right_detected=clear_to_disengaged, engaged=False
            ),
            EstopState.LOGICALLY_ENGAGED,
        ),
    ]

    for input, result in steps:
        subject.detector_listener(input)
        assert subject.state == result

    assert (
        subject.acknowledge_and_clear() == EstopState.DISENGAGED
        if clear_to_disengaged
        else EstopState.NOT_PRESENT
    )



import asyncio
from typing import AsyncGenerator

import pytest
from decoy import Decoy

from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control import ExecutionManager, modules
from opentrons.hardware_control.modules.tempdeck import TempDeck
from opentrons.hardware_control.modules.types import (
    ModuleDisconnectedCallback,
    ModuleErrorCallback,
)


@pytest.fixture
def usb_port() -> USBPort:
    """Token USB port."""
    return USBPort(
        name="",
        port_number=0,
        device_path="/dev/ot_module_sim_tempdeck0",
    )


@pytest.fixture
async def subject(
    usb_port: USBPort,
    mock_execution_manager: ExecutionManager,
    module_error_callback: ModuleErrorCallback,
    module_disconnected_callback: ModuleDisconnectedCallback,
) -> AsyncGenerator[modules.AbstractModule, None]:
    """Test subject"""
    temp = await modules.build(
        port="/dev/ot_module_sim_tempdeck0",
        usb_port=usb_port,
        type=modules.ModuleType["TEMPERATURE"],
        simulating=True,
        hw_control_loop=asyncio.get_running_loop(),
        execution_manager=mock_execution_manager,
        error_callback=module_error_callback,
        disconnected_callback=module_disconnected_callback,
    )
    try:
        yield temp
    finally:
        await temp.cleanup()


async def test_sim_initialization(subject: modules.AbstractModule) -> None:
    """It should build a tempdeck."""
    assert isinstance(subject, modules.AbstractModule)


async def test_sim_state(subject: modules.AbstractModule) -> None:
    """It sohuld forward state."""
    assert isinstance(subject, TempDeck)
    assert subject.temperature == 0
    assert subject.target is None
    assert subject.status == "idle"
    assert subject.live_data["status"] == subject.status

    live_data = subject.live_data["data"]
    assert modules.ModuleDataValidator.is_temperature_module_data(live_data)
    assert live_data["currentTemp"] == subject.temperature
    assert live_data["targetTemp"] == subject.target
    status = subject.device_info
    assert status["serial"] == "dummySerialTD"
    # return v1 if sim_model is not passed
    assert status["model"] == "temp_deck_v1.1"
    assert status["version"] == "dummyVersionTD"


async def test_sim_update(subject: modules.AbstractModule) -> None:
    """It should update state."""
    assert isinstance(subject, TempDeck)
    await subject.start_set_temperature(10)
    await subject.await_temperature(None)
    assert subject.temperature == 10
    assert subject.target == 10
    assert subject.status == "holding at target"
    await subject.deactivate()
    assert subject.temperature == 23
    assert subject.target is None
    assert subject.status == "idle"


async def test_revision_model_parsing(subject: modules.AbstractModule) -> None:
    """It should parse its model."""
    assert isinstance(subject, TempDeck)
    subject._device_info["model"] = "temp_deck_v20"
    assert subject.model() == "temperatureModuleV2"
    subject._device_info["model"] = "temp_deck_v4.0"
    assert subject.model() == "temperatureModuleV1"
    del subject._device_info["model"]
    assert subject.model() == "temperatureModuleV1"
    subject._device_info["model"] = "temp_deck_v1.1"
    assert subject.model() == "temperatureModuleV1"


async def test_error_callback(
    subject: modules.TempDeck,
    monkeypatch: pytest.MonkeyPatch,
    decoy: Decoy,
    module_error_callback: ModuleErrorCallback,
) -> None:
    """It should forward temperature check errors."""
    mock_get_temp = decoy.mock(func=subject._driver.get_temperature)
    exc = Exception("oh no!")
    decoy.when(await mock_get_temp()).then_raise(exc)
    monkeypatch.setattr(subject._driver, "get_temperature", mock_get_temp)
    with pytest.raises(Exception, match="oh no!"):
        await subject._poller.wait_next_poll()
    decoy.verify(
        module_error_callback(
            exc, "temperatureModuleV1", "/dev/ot_module_sim_tempdeck0", "dummySerialTD"
        )
    )



"""Tests for OT3 calibration."""

import copy
import json
from dataclasses import replace
from math import isclose
from typing import Any, AsyncIterator, Iterator, Literal, Tuple

import pytest
from mock import AsyncMock, Mock, patch
from mock import call as mock_call

from opentrons_shared_data.deck import get_calibration_square_position_in_slot
from opentrons_shared_data.errors.exceptions import (
    CalibrationStructureNotFoundError,
    EarlyCapacitiveSenseTrigger,
    EdgeNotFoundError,
    InaccurateNonContactSweepError,
)

from opentrons.config.types import OT3CalibrationSettings
from opentrons.hardware_control import ThreadManager
from opentrons.hardware_control.ot3_calibration import (
    EDGES,
    PREP_OFFSET_DEPTH,
    CalibrationMethod,
    _edges_from_data,
    _probe_deck_at,
    _verify_edge_pos,
    calibrate_pipette,
    find_axis_center,
    find_calibration_structure_height,
    find_edge_binary,
    find_slot_center_binary,
    find_slot_center_noncontact,
)
from opentrons.hardware_control.ot3api import OT3API
from opentrons.hardware_control.types import Axis, InstrumentProbeType, OT3Mount
from opentrons.types import Point


@pytest.fixture(autouse=True)
def mock_save_json() -> Iterator[Mock]:
    with patch("json.dump", Mock(spec=json.dump)) as jd:
        yield jd


@pytest.fixture
def mock_move_to(ot3_hardware: ThreadManager[OT3API]) -> Iterator[AsyncMock]:
    managed = ot3_hardware.managed_obj
    assert managed
    with patch.object(
        managed,
        "move_to",
        AsyncMock(
            spec=managed.move_to,
            wraps=managed.move_to,
        ),
    ) as mock_move:
        yield mock_move


@pytest.fixture
def mock_find_edge_binary() -> Iterator[AsyncMock]:
    with patch(
        "opentrons.hardware_control.ot3_calibration.find_edge_binary",
        AsyncMock(
            spec=find_edge_binary,
        ),
    ) as mock_find_edge_binary:
        yield mock_find_edge_binary


@pytest.fixture
def mock_capacitive_probe(ot3_hardware: ThreadManager[OT3API]) -> Iterator[AsyncMock]:
    managed = ot3_hardware.managed_obj
    assert managed
    with patch.object(
        managed,
        "capacitive_probe",
        AsyncMock(
            spec=managed.capacitive_probe,
            wraps=managed.capacitive_probe,
        ),
    ) as mock_probe:
        yield mock_probe


@pytest.fixture
def mock_probe_deck() -> Iterator[AsyncMock]:
    with patch(
        "opentrons.hardware_control.ot3_calibration._probe_deck_at",
        AsyncMock(
            spec=_probe_deck_at,
            wraps=_probe_deck_at,
        ),
    ) as mock_probe_deck:
        yield mock_probe_deck


@pytest.fixture
def mock_capacitive_sweep(ot3_hardware: ThreadManager[OT3API]) -> Iterator[AsyncMock]:
    managed = ot3_hardware.managed_obj
    assert managed
    with patch.object(
        managed,
        "capacitive_sweep",
        AsyncMock(
            spec=managed.capacitive_sweep,
            wraps=managed.capacitive_sweep,
        ),
    ) as mock_sweep:
        yield mock_sweep


@pytest.fixture
def mock_verify_edge(ot3_hardware: ThreadManager[OT3API]) -> Iterator[AsyncMock]:
    with patch(
        "opentrons.hardware_control.ot3_calibration._verify_edge_pos",
        AsyncMock(
            spec=_verify_edge_pos,
        ),
    ) as mock_verify_edge:
        yield mock_verify_edge


@pytest.fixture
def mock_data_analysis() -> Iterator[Mock]:
    with patch(
        "opentrons.hardware_control.ot3_calibration._edges_from_data",
        Mock(spec=_edges_from_data),
    ) as efd:
        yield efd


def _update_edge_sense_config(
    old: OT3CalibrationSettings, **new_edge_sense_settings: Any
) -> OT3CalibrationSettings:
    return replace(old, edge_sense=replace(old.edge_sense, **new_edge_sense_settings))


@pytest.fixture
async def override_cal_config(
    ot3_hardware: ThreadManager[OT3API],
) -> AsyncIterator[None]:
    old_calibration = copy.deepcopy(ot3_hardware.config.calibration)
    await ot3_hardware.update_config(
        calibration=_update_edge_sense_config(
            old_calibration,
            search_initial_tolerance_mm=8,
            search_iteration_limit=3,
            overrun_tolerance_mm=0,
            early_sense_tolerance_mm=2,
        )
    )
    try:
        yield
    finally:
        await ot3_hardware.update_config(calibration=old_calibration)


def _other_axis_val(point: Tuple[float, float, float], main_axis: Axis) -> float:
    if main_axis == Axis.X:
        return point[1]
    if main_axis == Axis.Y:
        return point[0]
    raise KeyError(main_axis)


# Mock Capacitive Probe Result (found_position[float], hit_deck[bool])
_HIT = (1, True)
_MISS = (-1, False)


@pytest.mark.parametrize(
    "direction_if_hit,probe_results,search_result",
    [
        # For each axis and direction, test
        # 1. hit-miss-miss
        # 2. miss-hit-hit
        # 3. miss-hit-miss
        (-1, (_HIT, _MISS, _MISS), -1),
        (-1, (_MISS, _HIT, _HIT), 1),
        (-1, (_MISS, _HIT, _MISS), 3),
        (1, (_HIT, _MISS, _MISS), 1),
        (1, (_MISS, _HIT, _HIT), -1),
        (1, (_MISS, _HIT, _MISS), -3),
    ],
)
async def test_find_edge(
    ot3_hardware: ThreadManager[OT3API],
    mock_capacitive_probe: AsyncMock,
    override_cal_config: None,
    mock_verify_edge: AsyncMock,
    mock_move_to: AsyncMock,
    direction_if_hit: Literal[1, -1],
    probe_results: Tuple[float, float, float],
    search_result: float,
) -> None:
    await ot3_hardware.home()
    mock_capacitive_probe.side_effect = probe_results
    edge_settings = ot3_hardware.config.calibration.edge_sense
    result = await find_edge_binary(
        ot3_hardware,
        OT3Mount.RIGHT,
        Point(0, 0, 0),
        Axis.X,
        edge_settings,
        direction_if_hit,
        False,
    )
    assert Axis.X.of_point(result) == search_result
    # the first move is in z only to the cal height
    checked_calls = mock_move_to.call_args_list[1:]
    # all other moves should only move in the search axis
    for call in checked_calls:
        assert call[0][0] == OT3Mount.RIGHT
        assert _other_axis_val(call[0][1], Axis.X) == pytest.approx(
            _other_axis_val(Point(0, 0, 0), Axis.X)
        )


async def test_find_slot_center(
    ot3_hardware: ThreadManager[OT3API], mock_find_edge_binary: AsyncMock
) -> None:
    await ot3_hardware.home()
    mock_find_edge_binary.return_value = Point(0, 0, 0)
    await find_slot_center_binary(
        ot3_hardware,
        OT3Mount.RIGHT,
        Point(0, 0, 0),
        False,
    )
    # we want to make sure that the final edge probe is at the same distance as the settings change
    final_pass_distance = 0.046875
    for call in mock_find_edge_binary.mock_calls:
        # call[1] is positional args and [4] is the EdgeSenseSettings arg
        settings = call[1][4]
        start_tolerance = settings.search_initial_tolerance_mm
        search_limit = settings.search_iteration_limit
        assert start_tolerance / (2**search_limit) == final_pass_distance


@pytest.mark.parametrize(
    "search_axis,direction_if_hit,probe_results",
    [
        (Axis.X, -1, (_HIT, _HIT)),
        (Axis.Y, -1, (_MISS, _MISS)),
    ],
)
async def test_edge_not_found(
    ot3_hardware: ThreadManager[OT3API],
    mock_capacitive_probe: AsyncMock,
    override_cal_config: None,
    mock_move_to: AsyncMock,
    search_axis: Literal[Axis.X, Axis.Y],
    direction_if_hit: Literal[1, -1],
    probe_results: Tuple[float, float, float],
) -> None:
    await ot3_hardware.home()
    mock_capacitive_probe.side_effect = probe_results
    with pytest.raises(EdgeNotFoundError):
        await _verify_edge_pos(
            ot3_hardware,
            OT3Mount.RIGHT,
            search_axis,
            Point(0, 0, 0),
            0.5,
            direction_if_hit,
        )


async def test_find_edge_early_trigger(
    ot3_hardware: ThreadManager[OT3API],
    mock_capacitive_probe: AsyncMock,
    override_cal_config: None,
) -> None:
    await ot3_hardware.home()
    mock_capacitive_probe.side_effect = ((3, True), ())
    edge_settings = ot3_hardware.config.calibration.edge_sense
    with pytest.raises(EarlyCapacitiveSenseTrigger):
        await find_edge_binary(
            ot3_hardware,
            OT3Mount.RIGHT,
            Point(0.0, 0.0, 0.0),
            Axis.Y,
            edge_settings,
            -1,
        )


async def test_deck_not_found(
    ot3_hardware: ThreadManager[OT3API],
    mock_capacitive_probe: AsyncMock,
    override_cal_config: None,
) -> None:
    await ot3_hardware.home()
    mock_capacitive_probe.side_effect = ((-25, False), ())
    with pytest.raises(CalibrationStructureNotFoundError):
        await find_calibration_structure_height(
            ot3_hardware,
            OT3Mount.RIGHT,
            Point(0.0, 0.0, 0.0),
        )


@pytest.mark.parametrize("mount", (OT3Mount.RIGHT, OT3Mount.LEFT))
@pytest.mark.parametrize("target", (Point(10, 10, 0), Point(355, 355, 0)))
async def test_find_deck_checks_z_only(
    ot3_hardware: ThreadManager[OT3API],
    mock_capacitive_probe: AsyncMock,
    override_cal_config: None,
    mock_probe_deck: AsyncMock,
    mock_move_to: AsyncMock,
    mount: OT3Mount,
    target: Point,
) -> None:
    await ot3_hardware.home()
    here = await ot3_hardware.gantry_position(mount)
    mock_capacitive_probe.side_effect = ((-1.8, True),)
    await find_calibration_structure_height(ot3_hardware, mount, target)

    z_prep_loc = target + PREP_OFFSET_DEPTH

    mock_probe_deck.assert_called_once_with(
        ot3_hardware,
        mount,
        z_prep_loc,
        ot3_hardware.config.calibration.z_offset.pass_settings,
        probe=InstrumentProbeType.PRIMARY,
    )
    # first we move only to safe height from current position
    first_move_point = mock_move_to.call_args_list[0][0][1]
    assert first_move_point.x == here.x
    assert first_move_point.y == here.y

    # actually move to the target position
    second_move_point = mock_move_to.call_args_list[1][0][1]
    assert isclose(second_move_point.x, z_prep_loc.x)
    assert isclose(second_move_point.y, z_prep_loc.y)


async def test_method_enum(
    ot3_hardware: ThreadManager[OT3API],
    override_cal_config: None,
) -> None:
    managed = ot3_hardware.managed_obj
    assert managed
    with (
        patch(
            "opentrons.hardware_control.ot3_calibration.find_slot_center_binary",
            AsyncMock(spec=find_slot_center_binary),
        ) as binary,
        patch(
            "opentrons.hardware_control.ot3_calibration.get_calibration_square_position_in_slot",
            Mock(),
        ) as calibration_target,
        patch(
            "opentrons.hardware_control.ot3_calibration.find_slot_center_noncontact",
            AsyncMock(spec=find_slot_center_noncontact),
        ) as noncontact,
        patch(
            "opentrons.hardware_control.ot3_calibration.find_calibration_structure_height",
            AsyncMock(spec=find_calibration_structure_height),
        ) as find_deck,
        patch.object(
            managed, "reset_instrument_offset", AsyncMock()
        ) as reset_instrument_offset,
        patch.object(
            managed, "save_instrument_offset", AsyncMock()
        ) as save_instrument_offset,
    ):
        find_deck.return_value = 10
        calibration_target.return_value = Point(0.0, 0.0, 0.0)
        binary.return_value = Point(1.0, 2.0, 3.0)
        noncontact.return_value = Point(3.0, 4.0, 5.0)
        await ot3_hardware.home()
        binval = await calibrate_pipette(
            ot3_hardware, OT3Mount.RIGHT, 5, CalibrationMethod.BINARY_SEARCH
        )
        reset_instrument_offset.assert_called_once()
        find_deck.assert_called_once()
        binary.assert_called_once()
        noncontact.assert_not_called()
        save_instrument_offset.assert_called_once()
        assert binval == Point(-1.0, -2.0, -3.0)

        reset_instrument_offset.reset_mock()
        find_deck.reset_mock()
        calibration_target.reset_mock()
        binary.reset_mock()
        noncontact.reset_mock()
        save_instrument_offset.reset_mock()

        ncval = await calibrate_pipette(
            ot3_hardware, OT3Mount.LEFT, 5, CalibrationMethod.NONCONTACT_PASS
        )
        reset_instrument_offset.assert_called_once()
        find_deck.assert_called_once()
        binary.assert_not_called()
        noncontact.assert_called_once()
        save_instrument_offset.assert_called_once()
        assert ncval == Point(-3.0, -4.0, -5.0)


async def test_calibrate_mount_errors(
    ot3_hardware: ThreadManager[OT3API], mock_data_analysis: Mock
) -> None:
    managed = ot3_hardware.managed_obj
    assert managed
    with (
        patch.object(
            managed, "reset_instrument_offset", AsyncMock()
        ) as reset_instrument_offset,
        patch.object(
            managed, "save_instrument_offset", AsyncMock()
        ) as save_instrument_offset,
        patch(
            "opentrons.hardware_control.ot3_calibration.find_calibration_structure_height",
            AsyncMock(spec=find_calibration_structure_height),
        ) as find_deck,
    ):
        find_deck.return_value = 10
        mock_data_analysis.return_value = (-1000, 1000)

        await ot3_hardware.home()
        # calibrate pipette should re-raise exception
        with pytest.raises(InaccurateNonContactSweepError):
            await calibrate_pipette(
                ot3_hardware, OT3Mount.RIGHT, 5, CalibrationMethod.NONCONTACT_PASS
            )

        reset_calls = [
            mock_call(OT3Mount.RIGHT),
            mock_call(OT3Mount.RIGHT, to_default=False),
        ]
        reset_instrument_offset.assert_has_calls(reset_calls)
        # instrument offset should not be saved
        save_instrument_offset.assert_not_called()

        reset_instrument_offset.reset_mock()
        save_instrument_offset.reset_mock()


async def test_noncontact_sanity(
    ot3_hardware: ThreadManager[OT3API],
    override_cal_config: None,
    mock_move_to: AsyncMock,
    mock_capacitive_sweep: AsyncMock,
    mock_data_analysis: Mock,
) -> None:
    mock_data_analysis.return_value = (-1000, 1000)
    await ot3_hardware.home()
    center = Point(*get_calibration_square_position_in_slot(5))
    with pytest.raises(InaccurateNonContactSweepError):
        await find_axis_center(
            ot3_hardware,
            OT3Mount.RIGHT,
            center + EDGES["left"],
            center + EDGES["right"],
            Axis.X,
        )
